##############################################################################
##
# This file is part of Sardana
##
# http://www.sardana-controls.org/
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""This is the demo macro module"""



__all__ = ["sar_demo", "clear_sar_demo", "sar_demo_hkl", "clear_sar_demo_hkl"]

import PyTango

from sardana.macroserver.macro import macro, Type
from sardana.macroserver.msexception import UnknownEnv

_ENV = "_SAR_DEMO"

_ENV_HKL = "_SAR_DEMO_HKL"


def get_free_names(db, prefix, nb, start_at=1):
    ret = []
    i = start_at
    failed = 96
    while len(ret) < nb and failed > 0:
        name = "%s%02d" % (prefix, i)
        try:
            db.get_device_alias(name)
            failed -= 1
        except:
            ret.append(name)
        i += 1
    if len(ret) < nb or failed == 0:
        raise Exception("Too many sardana demos registered on this system.\n"
                        "Please try using a different tango system")
    return ret


@macro()
def clear_sar_demo(self):
    """Undoes changes done with sar_demo"""
    try:
        SAR_DEMO = self.getEnv(_ENV)
    except:
        self.error("No demo has been prepared yet on this sardana!")
        return

    try:
        _ActiveMntGrp = self.getEnv("ActiveMntGrp")
    except UnknownEnv:
        _ActiveMntGrp = None

    self.print("Removing measurement groups...")
    for mg in SAR_DEMO.get("measurement_groups", ()):
        self.udefmeas(mg)
        if mg == _ActiveMntGrp:
            self.print("Unsetting ActiveMntGrp (was: %s)" % _ActiveMntGrp)
            self.unsetEnv("ActiveMntGrp")

    self.print("Removing elements...")
    elements = SAR_DEMO.get("elements", ())
    if len(elements) > 0:
        self.udefelem(elements)

    self.print("Removing controllers...")
    for ctrl in SAR_DEMO.get("controllers", ()):
        self.udefctrl(ctrl)

    self.print("Removing instruments...")
    pool = self.getPools()[0]
    for instrument in SAR_DEMO.get("instruments", ()):
        pool.DeleteElement(instrument)

    self.unsetEnv(_ENV)

    self.print("DONE!")


@macro()
def sar_demo(self):
    """Sets up a demo environment. It creates many elements for testing"""

    try:
        SAR_DEMO = self.getEnv(_ENV)
        self.error("A demo has already been prepared on this sardana")
        return
    except:
        pass

    db = PyTango.Database()

    mot_ctrl_name = get_free_names(db, "motctrl", 1)[0]
    ct_ctrl_name = get_free_names(db, "ctctrl", 1)[0]
    zerod_ctrl_name = get_free_names(db, "zerodctrl", 1)[0]
    oned_ctrl_name = get_free_names(db, "onedctrl", 1)[0]
    twod_ctrl_name = get_free_names(db, "twodctrl", 1)[0]
    tg_ctrl_name = get_free_names(db, "tgctrl", 1)[0]
    pm_ctrl_name = get_free_names(db, "slitctrl", 1)[0]
    ior_ctrl_name = get_free_names(db, "iorctrl", 1)[0]

    motor_names = get_free_names(db, "mot", 4)
    ct_names = get_free_names(db, "ct", 4)
    zerod_names = get_free_names(db, "zerod", 4)
    oned_names = get_free_names(db, "oned", 1)
    twod_names = get_free_names(db, "twod", 1)
    tg_names = get_free_names(db, "tg", 1)
    gap, offset = get_free_names(db, "gap", 1) + \
        get_free_names(db, "offset", 1)
    ior_names = get_free_names(db, "ior", 2)
    mg_name = get_free_names(db, "mntgrp", 1)[0]

    pools = self.getPools()
    if not len(pools):
        self.error('This is not a valid sardana demonstration system.\n'
                   'Sardana demonstration systems must be connect to at least '
                   'one Pool')
        return
    pool = pools[0]

    self.print("Creating motor controller", mot_ctrl_name, "...")
    self.defctrl("DummyMotorController", mot_ctrl_name)
    for axis, motor_name in enumerate(motor_names, 1):
        self.print("Creating motor", motor_name, "...")
        self.defelem(motor_name, mot_ctrl_name, axis)

    self.print("Creating counter controller", ct_ctrl_name, "...")
    self.defctrl("DummyCounterTimerController", ct_ctrl_name)
    for axis, ct_name in enumerate(ct_names, 1):
        self.print("Creating counter channel", ct_name, "...")
        self.defelem(ct_name, ct_ctrl_name, axis)

    self.print("Creating 0D controller", zerod_ctrl_name, "...")
    self.defctrl("DummyZeroDController", zerod_ctrl_name)
    for axis, zerod_name in enumerate(zerod_names, 1):
        self.print("Creating 0D channel", zerod_name, "...")
        self.defelem(zerod_name, zerod_ctrl_name, axis)

    self.print("Creating 1D controller", oned_ctrl_name, "...")
    self.defctrl("DummyOneDController", oned_ctrl_name)
    for axis, oned_name in enumerate(oned_names, 1):
        self.print("Creating 1D channel", oned_name, "...")
        self.defelem(oned_name, oned_ctrl_name, axis)

    self.print("Creating 2D controller", twod_ctrl_name, "...")
    self.defctrl("DummyTwoDController", twod_ctrl_name)
    for axis, twod_name in enumerate(twod_names, 1):
        self.print("Creating 2D channel", twod_name, "...")
        self.defelem(twod_name, twod_ctrl_name, axis)

    self.print("Creating Slit", pm_ctrl_name, "with", gap, ",", offset, "...")
    sl2t, sl2b = motor_names[:2]
    self.defctrl("Slit", pm_ctrl_name, ["sl2t=" + sl2t, "sl2b=" + sl2b,
                                        "Gap=" + gap, "Offset=" + offset])

    self.print("Creating trigger controller", tg_ctrl_name, "...")
    self.defctrl("DummyTriggerGateController", tg_ctrl_name)
    for axis, tg_name in enumerate(tg_names, 1):
        self.print("Creating trigger element", tg_name, "...")
        self.defelem(tg_name, tg_ctrl_name, axis)

    ct_ctrl = self.getController(ct_ctrl_name)
    ct_ctrl.getAttribute("synchronizer").write(tg_name)
    self.print("Connecting trigger/gate element", tg_name,
               "with counter/timer controller", ct_ctrl_name)

    self.print("Creating IORegister controller", ior_ctrl_name, "...")
    self.defctrl("DummyIORController", ior_ctrl_name)
    for axis, ior_name in enumerate(ior_names, 1):
        self.print("Creating IORegister", ior_name, "...")
        self.defelem(ior_name, ior_ctrl_name, axis)

    self.print("Creating measurement group", mg_name, "...")
    self.defmeas(mg_name, ct_names)

    try:
        self.getEnv("ActiveMntGrp")
    except UnknownEnv:
        self.print("Setting %s as ActiveMntGrp" % mg_name)
        self.setEnv("ActiveMntGrp", mg_name)

    self.print("Creating instruments: /slit, /mirror and /monitor ...")
    pool.createInstrument('/slit', 'NXcollimator')
    pool.createInstrument('/mirror', 'NXmirror')
    pool.createInstrument('/monitor', 'NXmonitor')

    self.print("Assigning elements to instruments...")
    self.getMotor(motor_names[0]).setInstrumentName('/slit')
    self.getMotor(motor_names[1]).setInstrumentName('/slit')
    self.getPseudoMotor(gap).setInstrumentName('/slit')
    self.getPseudoMotor(offset).setInstrumentName('/slit')
    self.getMotor(motor_names[2]).setInstrumentName('/mirror')
    self.getMotor(motor_names[3]).setInstrumentName('/mirror')
    self.getCounterTimer(ct_names[0]).setInstrumentName('/monitor')

    controllers = pm_ctrl_name, mot_ctrl_name, ct_ctrl_name, \
        zerod_ctrl_name, oned_ctrl_name, twod_ctrl_name, \
        tg_ctrl_name, ior_ctrl_name
    elements = [gap, offset] + motor_names + ct_names + \
        zerod_names + oned_names + twod_names + tg_names + \
        ior_names
    instruments = ["/slit", "/mirror", "/monitor"]
    d = dict(controllers=controllers, elements=elements,
             measurement_groups=[mg_name], instruments=instruments)

    self.setEnv(_ENV, d)

    self.print("DONE!")


@macro([["motor", Type.Moveable, None, '']])
def mym2(self, pm):
    self.output(pm.getMotorNames())
    elements = list(map(self.getMoveable, pm.elements))
    self.output(elements)
    self.output(type(pm))
    self.output(type(elements[0]))


@macro()
def clear_sar_demo_hkl(self):
    """Undoes changes done with sar_demo"""
    try:
        SAR_DEMO_HKL = self.getEnv(_ENV_HKL)
    except:
        self.error("No hkl demo has been prepared yet on this sardana!")
        return

    self.print("Removing hkl demo elements...")
    for elem in SAR_DEMO_HKL.get("elements", ()):
        self.udefelem(elem)

    self.print("Removing hkl demo controllers...")
    for ctrl in SAR_DEMO_HKL.get("controllers", ()):
        self.udefctrl(ctrl)

    self.unsetEnv(_ENV_HKL)

    self.clear_sar_demo()

    self.print("DONE!")


@macro()
def sar_demo_hkl(self):
    """Sets up a demo environment. It creates many elements for testing"""

    self.sar_demo()

    try:
        SAR_DEMO_HKL = self.getEnv(_ENV_HKL)
        self.error("An hkl demo has already been prepared on this sardana")
        return
    except:
        pass

    db = PyTango.Database()

    motor_ctrl_name = get_free_names(db, "motctrl", 1)[0]
    hkl_ctrl_name = get_free_names(db, "hklctrl", 1)[0]

    motor_names = []
    for motor in ["mu", "omega", "chi", "phi", "gamma", "delta"]:
        motor_names += get_free_names(db, motor, 1)

    pseudo_names = []
    for pseudo in ["h", "k", "l",
                   "psi",
                   "q", "alpha",
                   "qper", "qpar"]:
        pseudo_names += get_free_names(db, pseudo, 1)

    pools = self.getPools()
    if not len(pools):
        self.error('This is not a valid sardana demonstration system.\n'
                   'Sardana demonstration systems must be connect to at least '
                   'one Pool')
        return
    pool = pools[0]

    self.print("Creating motor controller", motor_ctrl_name, "...")
    self.defctrl("DummyMotorController", motor_ctrl_name)
    for axis, motor_name in enumerate(motor_names, 1):
        self.print("Creating motor", motor_name, "...")
        self.defelem(motor_name, motor_ctrl_name, axis)

    self.print("Creating hkl controller", hkl_ctrl_name, "...")
    self.defctrl("DiffracE6C", hkl_ctrl_name,
                 ["mu=" + motor_names[0],  # motor role
                  "omega=" + motor_names[1],
                  "chi=" + motor_names[2],
                  "phi=" + motor_names[3],
                  "gamma=" + motor_names[4],
                  "delta=" + motor_names[5],
                  "h=" + pseudo_names[0],  # pseudo role
                  "k=" + pseudo_names[1],
                  "l=" + pseudo_names[2],
                  "psi=" + pseudo_names[3],
                  "q=" + pseudo_names[4],
                  "alpha=" + pseudo_names[5],
                  "qper=" + pseudo_names[6],
                  "qpar=" + pseudo_names[7],
                  "diffractometertype", "E6C"])

    controllers = motor_ctrl_name, hkl_ctrl_name
    elements = pseudo_names + motor_names
    d = dict(controllers=controllers, elements=elements)
    self.setEnv(_ENV_HKL, d)

    self.print("DONE!")



""" This serves to upgrade MacroServer environment from Python 2 to Python 3:

IMPORTANT:
1. IT HAS TO BE USED WITH PYTHON 2.7!!!
2. CONDA python 2 package is missing the standard dbm package so this
   script will fail inside a conda python 2 environment
3. a new env file with an additional .db extension will be created. You
   should **NOT** change the macroserver EnvironmentDb property. The dbm
   will figure out automatically the file extension
4. a backup of the original environment will be available with the
   extension .py2
5. run the script with the same OS user as you run the Sardana/MacroServer
   server or after running the script give write permission to the newly
   created environment file to the OS user that you run the server

Usage: python upgrade_env.py <ms_dev_name|ms_dev_alias>
"""
import sys
import os
import shelve
import contextlib

import PyTango

assert sys.version_info[:2] in ((2, 6), (2, 7)),\
    "Must run with python 2.6 or 2.7"

DefaultEnvBaseDir = "/tmp/tango"
DefaultEnvRelDir = "%(ds_exec_name)s/%(ds_inst_name)s/macroserver.properties"


def get_ds_full_name(dev_name):
    db = PyTango.Database()
    db_dev = PyTango.DeviceProxy(db.dev_name())
    return db_dev.DbGetDeviceInfo(dev_name)[1][3]


def get_ms_properties(ms_name, ds_name):
    db = PyTango.Database()
    prop = db.get_device_property(ms_name, "EnvironmentDb")
    ms_properties = prop["EnvironmentDb"]
    if not ms_properties:
        dft_ms_properties = os.path.join(
            DefaultEnvBaseDir,
            DefaultEnvRelDir)
        ds_exec_name, ds_inst_name = ds_name.split("/")
        ms_properties = dft_ms_properties % {
            "ds_exec_name": ds_exec_name,
            "ds_inst_name": ds_inst_name}
    else:
        ms_properties = ms_properties[0]
    ms_properties = os.path.normpath(ms_properties)
    return ms_properties


def dbm_shelve(filename, backend, flag="c"):
    # NOTE: dbm appends '.db' to the end of the filename
    if backend == "gnu":
        import gdbm
        backend_dict = gdbm.open(filename, flag)
    elif backend == "dumb":
        import dumbdbm
        backend_dict = dumbdbm.open(filename, flag)
    return shelve.Shelf(backend_dict)


def migrate_file(filename, backend):
    assert not filename.endswith('.db'), \
        "Cannot migrate '.db' (It would be overwritten)"
    os.rename(filename, filename + '.py2')
    with contextlib.closing(shelve.open(filename + '.py2')) as src_db:
        data = dict(src_db)
    with contextlib.closing(dbm_shelve(filename, backend)) as dst_db:
        dst_db.update(data)


def upgrade_env(ms_name, backend):
    db = PyTango.Database()
    try:
        ms_info = db.get_device_info(ms_name)
        ds_name = ms_info.ds_full_name
    except AttributeError:
        ds_name = get_ds_full_name(ms_name)
    env_filename = get_ms_properties(ms_name, ds_name)
    migrate_file(env_filename, backend)


if __name__ == "__main__":
    if len(sys.argv) != 3:
        print("python upgrade_env.py <ms_dev_name> <gnu|dumb>")  # noqa
        sys.exit(1)
    ms_name = sys.argv[1]
    backend = sys.argv[2]
    if backend not in ("gnu", "dumb"):
        raise ValueError("{0} backend is not supported")
    upgrade_env(ms_name, backend)



#!/usr/bin/env python

##############################################################################
##
# This file is part of Sardana
##
# http://www.sardana-controls.org/
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

import operator
from enum import IntEnum

import PyTango

from taurus.external.qt import Qt, compat

import taurus
from taurus.core.util.colors import DEVICE_STATE_PALETTE
from taurus.core.taurusbasetypes import TaurusEventType
from taurus.core.tango.tangovalidator import TangoDeviceNameValidator
import taurus.qt.qtcore.mimetypes
from taurus.qt.qtgui.compact import TaurusReadWriteSwitcher
from taurus.qt.qtgui.dialog import ProtectTaurusMessageBox
from taurus.qt.qtgui.container import TaurusWidget
from taurus.qt.qtgui.container import TaurusFrame
from taurus.qt.qtgui.display import TaurusLabel
from taurus.qt.qtgui.input import TaurusValueLineEdit
from taurus.qt.qtgui.panel import DefaultLabelWidget
from taurus.qt.qtgui.panel import DefaultUnitsWidget
from taurus.qt.qtgui.panel import TaurusValue, TaurusAttrForm
from taurus.qt.qtcore.mimetypes import TAURUS_DEV_MIME_TYPE, TAURUS_ATTR_MIME_TYPE
from taurus.qt.qtgui.util.ui import UILoadable


class LimitsListener(Qt.QObject):
    """
    A class that listens to changes on motor limits.
    If that is the case it emits a signal so the application
    can do whatever with it.
    """

    updateLimits = Qt.pyqtSignal(compat.PY_OBJECT)

    def __init__(self):
        Qt.QObject.__init__(self)

    def eventReceived(self, evt_src, evt_type, evt_value):
        if evt_type not in [TaurusEventType.Change, TaurusEventType.Periodic]:
            return
        limits = evt_value.value
        self.updateLimits.emit(limits.tolist())


class Limit(IntEnum):
    Home = 0
    Positive = 1
    Negative = 2


class LimitInfo:

    def __init__(self, exist=False, active=False, status=""):
        # for software limit True when defined
        # for hardware limit True when physical motor, False when pseudo
        self.exist = exist
        self.active = active
        self.status = status


def eval_hw_limit(limit, value):
    """Evaluate hardware limit information

    :param limit: which limit (negative or positive)
    :type limit: `Limit`
    :param value: limit value, `None` means no hardware limit (pseudo motor)
    :type value: `bool` or `None`
    :return: limit information
    :rtype: `LimitInfo`
    """
    if limit is Limit.Positive:
        status = "Positive "
    elif limit is Limit.Negative:
        status = "Negative "
    else:
        raise ValueError("{} wrong limit".format(limit))
    exist = True
    active = False
    status += "hardware limit "
    if value is None:
        exist = False
        status += "does not exist."
    elif value:
        active = True
        status += "is active."
    else:
        status += "is not active."
    return LimitInfo(exist, active, status)


def eval_sw_limit(limit, position_attr):
    """Evaluate software limit information

    :param limit: which limit (negative or positive)
    :type limit: `Limit`
    :param position_attr: position attribute
    :type position_attr: `taurus.core.tango.TangoAttribute`
    :return: limit information
    :rtype: `LimitInfo`
    """
    if limit == Limit.Positive:
        getter_name = "getMaxRange"
        comparator = operator.ge
        not_defined = float("inf")
        status = "Positive "
    elif limit == Limit.Negative:
        getter_name = "getMinRange"
        comparator = operator.le
        not_defined = float("-inf")
        status = "Negative "
    else:
        raise ValueError("{} wrong limit".format(limit))
    active = False
    exist = False
    lim_value = getattr(position_attr, getter_name)().magnitude
    status += "software limit is "
    if lim_value == not_defined:
        status += "not defined."
    else:
        exist = True
        position = position_attr.read().rvalue.magnitude
        if comparator(position, lim_value):
            status += "active."
            active = True
        else:
            status += "defined."
    return LimitInfo(exist, active, status)


class PoolMotorClient():

    maxint_in_32_bits = 2147483647

    def __init__(self):
        self.motor_dev = None
        self.has_limits = False
        self.has_encoder = False

    def setMotor(self, pool_motor_dev_name):
        # AT SOME POINT THIS WILL BE USING THE 'POOL' TAURUS EXTENSION
        # TO OPERATE THE MOTOR INSTEAD OF A 'TANGO' TAURUSDEVICE
        try:
            self.motor_dev = taurus.Device(pool_motor_dev_name)
            # IT IS IMPORTANT TO KNOW IF IT IS AN ICEPAP MOTOR, SO EXTRA FEATURES CAN BE PROVIDED
            # PENDING.
            self.has_limits = hasattr(self.motor_dev, 'Limit_Switches')
            self.has_encoder = hasattr(self.motor_dev, 'Encoder')
        except Exception as e:
            taurus.warning('Exception Creating Motor Device %s', str(e))

    def moveMotor(self, pos):
        #self.motor_dev['position'] = pos
        # Make use of Taurus operations (being logged)
        self.motor_dev.getAttribute('Position').write(pos)

    def moveInc(self, inc):
        self.moveMotor(self.motor_dev['position'].value + inc)

    @Qt.pyqtSlot()
    def jogNeg(self):
        neg_limit = -((self.maxint_in_32_bits // 2) - 1)
        # THERE IS A BUG IN THE ICEPAP THAT DOES NOT ALLOW MOVE ABSOLUTE FURTHER THAN 32 BIT
        # SO IF THERE ARE STEPS PER UNIT, max_int HAS TO BE REDUCED
        if hasattr(self.motor_dev, 'step_per_unit'):
            neg_limit = neg_limit / self.motor_dev['step_per_unit'].value
        try:
            min_value = self.motor_dev.getAttribute(
                'Position').getConfig().getValueObj().min_value
            neg_limit = float(min_value)
        except Exception:
            pass
        self.moveMotor(neg_limit)

    @Qt.pyqtSlot()
    def jogPos(self):
        pos_limit = (self.maxint_in_32_bits // 2) - 1
        # THERE IS A BUG IN THE ICEPAP THAT DOES NOT ALLOW MOVE ABSOLUTE FURTHER THAN 32 BIT
        # SO IF THERE ARE STEPS PER UNIT, max_int HAS TO BE REDUCED
        if hasattr(self.motor_dev, 'step_per_unit'):
            pos_limit = pos_limit / self.motor_dev['step_per_unit'].value
        try:
            max_value = self.motor_dev.getAttribute(
                'Position').getConfig().getValueObj().max_value
            pos_limit = float(max_value)
        except Exception:
            pass
        self.moveMotor(pos_limit)

    @Qt.pyqtSlot()
    def goHome(self):
        pass

    @Qt.pyqtSlot()
    def abort(self):
        self.motor_dev.abort()


class LabelWidgetDragsDeviceAndAttribute(DefaultLabelWidget):
    """ Offer richer mime data with taurus-device, taurus-attribute, and plain-text. """

    def mouseMoveEvent(self, event):
        model = self.taurusValueBuddy().getModelName().encode('utf-8')
        mimeData = Qt.QMimeData()
        mimeData.setText(self.text())
        attr_name = model
        dev_name = model.rpartition(b'/')[0]
        mimeData.setData(TAURUS_DEV_MIME_TYPE, dev_name)
        mimeData.setData(TAURUS_ATTR_MIME_TYPE, attr_name)

        drag = Qt.QDrag(self)
        drag.setMimeData(mimeData)
        drag.setHotSpot(event.pos() - self.rect().topLeft())
        drag.exec_(Qt.Qt.CopyAction, Qt.Qt.CopyAction)


class PoolMotorConfigurationForm(TaurusAttrForm):

    def __init__(self, parent=None, designMode=False):
        TaurusAttrForm.__init__(self, parent, designMode)
        self._form.setWithButtons(False)

    def getMotorControllerType(self):
        modelObj = self.getModelObj()
        modelNormalName = modelObj.getNormalName()
        poolDsId = modelObj.getDeviceProxy().info().server_id
        db = taurus.Database()
        pool_devices = tuple(db.get_device_class_list(poolDsId).value_string)
        pool_dev_name = pool_devices[pool_devices.index('Pool') - 1]
        pool = taurus.Device(pool_dev_name)
        poolMotorInfos = pool["MotorList"].value
        for motorInfo in poolMotorInfos:
            # BE CAREFUL, THIS ONLY WORKS IF NOBODY CHANGES THE DEVICE NAME OF A MOTOR!!!
            # ALSO THERE COULD BE A CASE PROBLEM, BETTER DO COMPARISONS WITH .lower()
            # to better understand following actions
            # this is an example of one motor info record
            #'dummymotor10 (motor/dummymotorctrl/10) (dummymotorctrl/10) Motor',
            motorInfos = motorInfo.split()
            if modelNormalName.lower() == motorInfos[1][1:-1].lower():
                controllerName = motorInfos[2][1:-1].split("/")[0]

        poolControllerInfos = pool["ControllerList"].value
        for controllerInfo in poolControllerInfos:
            # to better understand following actions
            # this is an example of one controller info record
            #'dummymotorctrl (DummyMotorController.DummyMotorController/dummymotorctrl) - Motor Python ctrl (DummyMotorController.py)'
            controllerInfos = controllerInfo.split()
            if controllerName.lower() == controllerInfos[0].lower():
                controllerType = controllerInfos[1][1:-1].split("/")[0]
        return controllerType

    def getDisplayAttributes(self, controllerType):
        attributes = ['position',
                      'state',
                      'status',
                      'velocity',
                      'acceleration',
                      'base_rate',
                      'step_per_unit',
                      'dialposition',
                      'sign',
                      'offset',
                      'backlash']

        if controllerType == "IcePAPCtrl.IcepapController":
            attributes.insert(1, "encoder")
            attributes.extend(['frequency',
                               'poweron',
                               'closedloop',
                               'useencodersource',
                               'encodersource',
                               'encodersourceformula',
                               'statusstopcode',
                               'statusdisable',
                               'statusready',
                               'statuslim-',
                               'statuslim+',
                               'statushome'])

        elif controllerType == "PmacCtrl.PmacController":
            attributes.extend(["motoractivated",
                               "negativeendlimitset",
                               "positiveendlimitset",
                               "handwheelenabled",
                               "phasedmotor",
                               "openloopmode",
                               "runningdefine-timemove",
                               "integrationmode",
                               "dwellinprogress",
                               "datablockerror",
                               "desiredvelocityzero",
                               "abortdeceleration",
                               "blockrequest",
                               "homesearchinprogress",
                               "assignedtocoordinatesystem",
                               "coordinatesystem",
                               "amplifierenabled",
                               "stoppedonpositionlimit",
                               "homecomplete",
                               "phasingsearcherror",
                               "triggermove",
                               "integratedfatalfollowingerror",
                               "i2t_amplifierfaulterror",
                               "backlashdirectionflag",
                               "amplifierfaulterror",
                               "fatalfollowingerror",
                               "warningfollowingerror",
                               "inposition",
                               "motionprogramrunning"])

        elif controllerType == "TurboPmacCtrl.TurboPmacController":
            attributes.extend(["motoractivated",
                               "negativeendlimitset",
                               "positiveendlimitset",
                               "extendedservoalgorithmenabled"
                               "amplifierenabled",
                               "openloopmode",
                               "movetimeractive",
                               "integrationmode",
                               "dwellinprogress",
                               "datablockerror",
                               "desiredvelocityzero",
                               "abortdeceleration",
                               "blockrequest",
                               "homesearchinprogress",
                               "user-writtenphaseenable",
                               "user-writtenservoenable",
                               "alternatesource/destination",
                               "phasedmotor",
                               "followingoffsetmode",
                               "followingenabled",
                               "errortriger",
                               "softwarepositioncapture",
                               "integratorinvelocityloop",
                               "alternatecommand-outputmode",
                               "coordinatesystem",
                               "coordinatedefinition",
                               "assignedtocoordinatesystem",
                               "foregroundinposition",
                               "stoppedondesiredpositionlimit",
                               "stoppedonpositionlimit",
                               "homecomplete",
                               "phasing_search/read_active",
                               "triggermove",
                               "integratedfatalfollowingerror",
                               "i2t_amplifierfaulterror",
                               "backlashdirectionflag",
                               "amplifierfaulterror",
                               "fatalfollowingerror",
                               "warningfollowingerror",
                               "inposition"])
        return attributes

    def setModel(self, modelName):
        TaurusAttrForm.setModel(self, modelName)
        controllerType = self.getMotorControllerType()
        attributes = self.getDisplayAttributes(controllerType)
        #self.setViewFilters([lambda a: a.name.lower() in attributes])
        self.setSortKey(lambda att: attributes.index(
            att.name.lower()) if att.name.lower() in attributes else 1)


##########################################################################
# NEW APPROACH TO OPERATE POOL MOTORS FROM A TAURUS FORM INHERITTING DIRECTLY FROM TaurusVALUE #
# AND USING PARTICULAR CLASSES THAT KNOW THEY ARE PART OF A TAURUSVALUE AND CAN INTERACT       #
##########################################################################

class TaurusAttributeListener(Qt.QObject):
    """
    A class that recieves events on tango attribute changes.
    If that is the case it emits a signal with the event's value.
    """

    eventReceivedSignal = Qt.pyqtSignal(compat.PY_OBJECT)

    def __init__(self):
        Qt.QObject.__init__(self)

    def eventReceived(self, evt_src, evt_type, evt_value):
        if evt_type not in [TaurusEventType.Change, TaurusEventType.Periodic]:
            return
        try:
            value = evt_value.rvalue.magnitude
        except AttributeError:
            # spectrum/images or str attribute values are not Quantities
            value = evt_value.rvalue
        self.eventReceivedSignal.emit(value)


##################################################
#                  LABEL WIDGET                  #
##################################################
class PoolMotorTVLabelWidget(TaurusWidget):
    '''
    @TODO tooltip should be extended with status info
    @TODO context menu should be the lbl_alias extended
    @TODO default tooltip extended with the complete (multiline) status
    @TODO rightclick popup menu with actions: (1) switch user/expert view, (2) Config -all attributes-, (3) change motor
    For the (3), a drop event should accept if it is a device, and add it to the 'change-motor' list and select
    @TODO on the 'expert' row, it could be an ENABLE section with a button to set PowerOn to True/False
    '''

    layoutAlignment = Qt.Qt.AlignTop

    def __init__(self, parent=None, designMode=False):
        TaurusWidget.__init__(self, parent, designMode)
        self.setLayout(Qt.QGridLayout())
        self.layout().setContentsMargins(0, 0, 0, 0)
        self.layout().setSpacing(0)

        self.lbl_alias = DefaultLabelWidget(parent, designMode)
        self.lbl_alias.setBgRole('none')
        self.layout().addWidget(self.lbl_alias)

        self.btn_poweron = Qt.QPushButton()
        self.btn_poweron.setText('Set ON')
        self.layout().addWidget(self.btn_poweron)

        # Align everything on top
        self.layout().addItem(Qt.QSpacerItem(
            1, 1, Qt.QSizePolicy.Minimum, Qt.QSizePolicy.Expanding))

        # I don't like this approach, there should be something like
        # self.lbl_alias.addAction(...)
        self.lbl_alias.contextMenuEvent = \
            lambda event: self.contextMenuEvent(event)

        # I' don't like this approach, there should be something like
        # self.lbl_alias.addToolTipCallback(self.calculate_extra_tooltip)
        self.lbl_alias.getFormatedToolTip = self.calculateExtendedTooltip

        # I' don't like this approach, there should be something like
        # self.lbl_alias.disableDrag() or self.lbl_alias.setDragEnabled(False)
        # or better, define if Attribute or Device or Both have to be included
        # in the mimeData
        self.lbl_alias.mouseMoveEvent = self.mouseMoveEvent

    def setExpertView(self, expertView):
        btn_poweron_visible = expertView \
                              and self.taurusValueBuddy().hasPowerOn()
        self.btn_poweron.setVisible(btn_poweron_visible)

    @Qt.pyqtSlot()
    @ProtectTaurusMessageBox(
        msg='An error occurred trying to write PowerOn Attribute.')
    def setPowerOn(self):
        motor_dev = self.taurusValueBuddy().motor_dev
        if motor_dev is not None:
            poweron = (self.btn_poweron.text() == 'Set ON')
            motor_dev.getAttribute('PowerOn').write(poweron)

    def setModel(self, model):
        # Handle User/Expert view
        try:
            self.taurusValueBuddy().expertViewChanged.disconnect(
                self.setExpertView)
        except TypeError:
            pass
        try:
            self.btn_poweron.clicked.disconnect(self.setPowerOn)
        except TypeError:
            pass
        if model in (None, ''):
            self.lbl_alias.setModel(model)
            TaurusWidget.setModel(self, model)
            return
        self.lbl_alias.taurusValueBuddy = self.taurusValueBuddy
        self.lbl_alias.setModel(model)
        TaurusWidget.setModel(self, model + '/Status')
        self.taurusValueBuddy().expertViewChanged.connect(
            self.setExpertView)
        # Handle Power ON/OFF
        self.btn_poweron.clicked.connect(self.setPowerOn)
        self.setExpertView(self.taurusValueBuddy()._expertView)

    def calculateExtendedTooltip(self, cache=False):
        default_label_widget_tooltip = DefaultLabelWidget.getFormatedToolTip(
            self.lbl_alias, cache)
        status_info = ''
        motor_dev = self.taurusValueBuddy().motor_dev
        if motor_dev is not None:
            status = motor_dev.getAttribute('Status').read().rvalue
            # MAKE IT LOOK LIKE THE STANDARD TABLE FOR TAURUS TOOLTIPS
            status_lines = status.split('\n')
            status_info = '<TABLE width="500" border="0" cellpadding="1" cellspacing="0"><TR><TD WIDTH="80" ALIGN="RIGHT" VALIGN="MIDDLE"><B>Status:</B></TD><TD>' + \
                status_lines[0] + '</TD></TR>'
            for status_extra_line in status_lines[1:]:
                status_info += '<TR><TD></TD><TD>' + status_extra_line + '</TD></TR>'
            status_info += '</TABLE>'
        return default_label_widget_tooltip + status_info

    def contextMenuEvent(self, event):
        # Overwrite the default taurus label behaviour
        menu = Qt.QMenu(self)
        action_expert_view = Qt.QAction(self)
        action_expert_view.setText('Expert View')
        action_expert_view.setCheckable(True)
        action_expert_view.setChecked(self.taurusValueBuddy()._expertView)
        menu.addAction(action_expert_view)
        action_expert_view.toggled.connect(
            self.taurusValueBuddy().setExpertView)

        action_tango_attributes = Qt.QAction(self)
        action_tango_attributes.setIcon(
            Qt.QIcon("categories:preferences-system.svg"))
        action_tango_attributes.setText('Tango Attributes')
        menu.addAction(action_tango_attributes)
        action_tango_attributes.triggered.connect(
            self.taurusValueBuddy().showTangoAttributes)

        cm_action = menu.addAction("Compact")
        cm_action.setCheckable(True)
        cm_action.setChecked(self.taurusValueBuddy().isCompact())
        cm_action.toggled.connect(self.taurusValueBuddy().setCompact)

        menu.exec_(event.globalPos())
        event.accept()

    def mouseMoveEvent(self, event):
        model = self.taurusValueBuddy().getModelObj()
        mimeData = Qt.QMimeData()
        mimeData.setText(self.lbl_alias.text())
        dev_name = model.getFullName().encode('utf-8')
        attr_name = dev_name + b'/Position'
        mimeData.setData(TAURUS_DEV_MIME_TYPE, dev_name)
        mimeData.setData(TAURUS_ATTR_MIME_TYPE, attr_name)

        drag = Qt.QDrag(self)
        drag.setMimeData(mimeData)
        drag.setHotSpot(event.pos() - self.rect().topLeft())
        drag.exec_(Qt.Qt.CopyAction, Qt.Qt.CopyAction)

##################################################
#                   READ WIDGET                  #
##################################################


class PoolMotorTVReadWidget(TaurusWidget):
    '''
    @TODO on the 'expert' row, there should be an Indexer/Encoder radiobuttongroup to show units from raw dial/indx/enc
    @TODO TaurusLCD may be used but, now it does not display the sign, and color is WHITE...
    '''

    layoutAlignment = Qt.Qt.AlignTop

    def __init__(self, parent=None, designMode=False):
        TaurusWidget.__init__(self, parent, designMode)

        self.setLayout(Qt.QGridLayout())
        self.layout().setContentsMargins(0, 0, 0, 0)
        self.layout().setSpacing(0)

        limits_layout = Qt.QHBoxLayout()
        limits_layout.setContentsMargins(0, 0, 0, 0)
        limits_layout.setSpacing(0)

        self.lim_neg = Qt.QLabel()
        self.prepare_limit(self.lim_neg)
        self._config_limit(Limit.Negative,
                           hw_lim=LimitInfo(),
                           sw_lim=LimitInfo())
        limits_layout.addWidget(self.lim_neg)

        self.lim_pos = Qt.QLabel()
        self.prepare_limit(self.lim_pos)
        self._config_limit(Limit.Positive,
                           hw_lim=LimitInfo(),
                           sw_lim=LimitInfo())
        limits_layout.addWidget(self.lim_pos)

        self.layout().addLayout(limits_layout, 0, 0)

        self.lbl_read = TaurusLabel()
        self.lbl_read.setBgRole('quality')
        self.lbl_read.setSizePolicy(Qt.QSizePolicy(
            Qt.QSizePolicy.Expanding, Qt.QSizePolicy.Fixed))
        self.layout().addWidget(self.lbl_read, 0, 1)

        # WITH A COMPACT VIEW, BETTER TO BE ABLE TO STOP!
        self.btn_stop = Qt.QPushButton()
        self.btn_stop.setToolTip('Stops the motor')
        self.prepare_button(self.btn_stop)
        self.btn_stop.setIcon(Qt.QIcon("actions:media_playback_stop.svg"))
        self.layout().addWidget(self.btn_stop, 0, 2)

        self.btn_stop.clicked.connect(self.abort)

        # WITH COMPACT VIEW, WE NEED TO FORWARD DOUBLE CLICK EVENT
        self.lbl_read.installEventFilter(self)

        # @TODO right now, no options here...
        #self.cb_expertRead = Qt.QComboBox()
        # self.cb_expertRead.addItems(['Enc'])
        #self.layout().addWidget(self.cb_expertRead, 1, 0)

        self.lbl_enc = None
        self.lbl_enc_read = None

        # Align everything on top
        self.layout().addItem(Qt.QSpacerItem(
            1, 1, Qt.QSizePolicy.Minimum, Qt.QSizePolicy.Expanding), 2, 0, 1, 2)

        # IN ORDER TO BEHAVE AS EXPECTED REGARDING THE 'COMPACT VIEW' FEATURE
        # WE NEED TO SET THE 'EXPERTVIEW' WITHOUT ACCESSING THE taurusValueBuddy WHICH IS STILL NOT LINKED
        # SO WE ASSUME 'expertview is FALSE' AND WE HAVE TO AVOID self.setExpertView :-(
        # WOULD BE NICE THAT THE taurusValueBuddy COULD EMIT THE PROPER
        # SIGNAL...

    def eventFilter(self, obj, event):
        if event.type() == Qt.QEvent.MouseButtonDblClick:
            if isinstance(self.parent(), TaurusReadWriteSwitcher):
                self.parent().enterEdit()
                return True
        try:
            if obj is self.lbl_read:
                return self.lbl_read.eventFilter(obj, event)
        except AttributeError:
            # self.lbl_read may not exist now
            pass
        return True

    @Qt.pyqtSlot()
    @ProtectTaurusMessageBox(msg='An error occurred trying to abort the motion.')
    def abort(self):
        motor_dev = self.taurusValueBuddy().motor_dev
        if motor_dev is not None:
            motor_dev.abort()

    def setExpertView(self, expertView):
        if self.lbl_enc_read is None:
            return
        self.lbl_enc.setVisible(False)
        self.lbl_enc_read.setVisible(False)
        if expertView and self.taurusValueBuddy().motor_dev is not None:
            encoder = self.taurusValueBuddy().hasEncoder()
            self.lbl_enc.setVisible(encoder)
            self.lbl_enc_read.setVisible(encoder)

    def prepare_button(self, btn):
        btn_policy = Qt.QSizePolicy(Qt.QSizePolicy.Fixed, Qt.QSizePolicy.Fixed)
        btn_policy.setHorizontalStretch(0)
        btn_policy.setVerticalStretch(0)
        btn.setSizePolicy(btn_policy)
        btn.setMinimumSize(25, 25)
        btn.setMaximumSize(25, 25)
        btn.setText('')

    def prepare_limit(self, lbl):
        lbl_policy = Qt.QSizePolicy(Qt.QSizePolicy.Fixed, Qt.QSizePolicy.Fixed)
        lbl_policy.setHorizontalStretch(0)
        lbl_policy.setVerticalStretch(0)
        lbl.setSizePolicy(lbl_policy)
        lbl.setText('')

    def _config_limit(self, limit, hw_lim, sw_lim):
        """Configure sub-widgets according to limit information

        :param limit: which limit (negative or positive)
        :type limit: `Limit`
        :param hw_lim: hardware limit information
        :type hw_lim: `LimitInfo`
        :param hw_lim: software limit information
        :type hw_lim: `LimitInfo`
        """
        if limit == Limit.Negative:
            widget = self.lim_neg
            icon = Qt.QIcon("designer:minus.png")
        elif limit == Limit.Positive:
            widget = self.lim_pos
            icon = Qt.QIcon("designer:plus.png")
        else:
            raise ValueError("{} wrong limit".format(limit))
        widget.setStyleSheet('')
        if hw_lim.exist or sw_lim.exist:
            pixmap = Qt.QPixmap(icon.pixmap(Qt.QSize(32, 32),
                                            Qt.QIcon.Active))
            if hw_lim.active or sw_lim.active:
                colour = DEVICE_STATE_PALETTE.qtStyleSheet(
                    PyTango.DevState.ALARM)
                widget.setStyleSheet(colour)
        else:
            pixmap = Qt.QPixmap(icon.pixmap(Qt.QSize(32, 32),
                                            Qt.QIcon.Disabled))
        tool_tip = hw_lim.status + " " + sw_lim.status
        widget.setToolTip(tool_tip)
        widget.setPixmap(pixmap)

    def _create_encoder(self):
        if self.taurusValueBuddy().hasEncoder() and self.lbl_enc_read is None:
            self.lbl_enc = Qt.QLabel('Encoder')
            self.layout().addWidget(self.lbl_enc, 1, 0)

            self.lbl_enc_read = TaurusLabel()
            self.lbl_enc_read.setBgRole('none')
            self.lbl_enc_read.setSizePolicy(Qt.QSizePolicy(
                Qt.QSizePolicy.Expanding, Qt.QSizePolicy.Fixed))
            self.layout().addWidget(self.lbl_enc_read, 1, 1)

            # Align everything on top
            self.layout().addItem(Qt.QSpacerItem(
                1, 1, Qt.QSizePolicy.Minimum,
                Qt.QSizePolicy.Expanding), 2, 0, 1, 2)

            self.lbl_enc.setVisible(False)
            self.lbl_enc_read.setVisible(False)

    def setModel(self, model):
        if hasattr(self, 'taurusValueBuddy'):
            try:
                self.taurusValueBuddy().expertViewChanged.disconnect(
                    self.setExpertView)
            except TypeError:
                pass
        if model in (None, ''):
            TaurusWidget.setModel(self, model)
            self.lbl_read.setModel(model)
            if self.lbl_enc_read is not None:
                self.lbl_enc_read.setModel(model)
            return
        TaurusWidget.setModel(self, model + '/Position')
        self.lbl_read.setModel(model + '/Position')
        if (self.taurusValueBuddy().hasEncoder()
                and self.lbl_enc_read is None):
            self._create_encoder()
            self.lbl_enc_read.setModel(model + '/Encoder')
        # Handle User/Expert view
        self.setExpertView(self.taurusValueBuddy()._expertView)
        self.taurusValueBuddy().expertViewChanged.connect(
            self.setExpertView)

##################################################
#                  WRITE WIDGET                  #
##################################################

class PoolMotorTVWriteWidget(TaurusWidget):

    layoutAlignment = Qt.Qt.AlignTop

    applied = Qt.pyqtSignal()

    def __init__(self, parent=None, designMode=False):
        TaurusWidget.__init__(self, parent, designMode)

        # ------------------------------------------------------------
        # Workaround for Taurus3 support
        if int(taurus.Release.version.split('.')[0]) < 4:
            from taurus.qt.qtgui.base.taurusbase import baseOldSignal
            self.applied = baseOldSignal('applied', self)
        # ------------------------------------------------------------

        self.setLayout(Qt.QGridLayout())
        self.layout().setContentsMargins(0, 0, 0, 0)
        self.layout().setSpacing(0)

        self.le_write_absolute = TaurusValueLineEdit()
        self.layout().addWidget(self.le_write_absolute, 0, 0)

        self.qw_write_relative = Qt.QWidget()
        self.qw_write_relative.setLayout(Qt.QHBoxLayout())
        self.qw_write_relative.layout().setContentsMargins(0, 0, 0, 0)
        self.qw_write_relative.layout().setSpacing(0)

        self.cb_step = Qt.QComboBox()
        self.cb_step.setSizePolicy(Qt.QSizePolicy(
            Qt.QSizePolicy.Expanding, Qt.QSizePolicy.Fixed))
        self.cb_step.setEditable(True)
        self.cb_step.lineEdit().setValidator(Qt.QDoubleValidator(self))
        self.cb_step.lineEdit().setAlignment(Qt.Qt.AlignRight)
        self.cb_step.addItem('1')
        self.qw_write_relative.layout().addWidget(self.cb_step)

        self.btn_step_down = Qt.QPushButton()
        self.btn_step_down.setToolTip('Decrements motor position')
        self.prepare_button(self.btn_step_down)
        self.btn_step_down.setIcon(
            Qt.QIcon("actions:media_playback_backward.svg"))
        self.qw_write_relative.layout().addWidget(self.btn_step_down)

        self.btn_step_up = Qt.QPushButton()
        self.btn_step_up.setToolTip('Increments motor position')
        self.prepare_button(self.btn_step_up)
        self.btn_step_up.setIcon(Qt.QIcon("actions:media_playback_start.svg"))
        self.qw_write_relative.layout().addWidget(self.btn_step_up)

        self.layout().addWidget(self.qw_write_relative, 0, 0)

        self.cbAbsoluteRelative = Qt.QComboBox()
        self.cbAbsoluteRelative.currentIndexChanged['QString'].connect(
            self.cbAbsoluteRelativeChanged)
        self.cbAbsoluteRelative.addItems(['Abs', 'Rel'])
        self.layout().addWidget(self.cbAbsoluteRelative, 0, 1)
        self.registerConfigProperty(
            self.cbAbsoluteRelative.currentIndex,
            self.cbAbsoluteRelative.setCurrentIndex, 'AbsRelIndex')
        # WITH THE COMPACCT VIEW FEATURE, BETTER TO HAVE IT IN THE READ WIDGET
        # WOULD BE BETTER AS AN 'EXTRA WIDGET' (SOME DAY...)
        #self.btn_stop = Qt.QPushButton()
        #self.btn_stop.setToolTip('Stops the motor')
        # self.prepare_button(self.btn_stop)
        # self.btn_stop.setIcon(getIcon(':/actions/media_playback_stop.svg'))
        #self.layout().addWidget(self.btn_stop, 0, 2)

        btns_layout = Qt.QHBoxLayout()
        btns_layout.setContentsMargins(0, 0, 0, 0)
        btns_layout.setSpacing(0)

        btns_layout.addItem(Qt.QSpacerItem(
            1, 1, Qt.QSizePolicy.Expanding, Qt.QSizePolicy.Minimum))

        self.btn_to_neg = Qt.QPushButton()
        self.btn_to_neg.setToolTip(
            'Moves the motor towards the Negative Software Limit')
        self.prepare_button(self.btn_to_neg)
        self.btn_to_neg.setIcon(Qt.QIcon("actions:media_skip_backward.svg"))
        btns_layout.addWidget(self.btn_to_neg)

        self.btn_to_neg_press = Qt.QPushButton()
        self.btn_to_neg_press.setToolTip(
            'Moves the motor (while pressed) towards the Negative Software Limit')
        self.prepare_button(self.btn_to_neg_press)
        self.btn_to_neg_press.setIcon(
            Qt.QIcon("actions:media_seek_backward.svg"))
        btns_layout.addWidget(self.btn_to_neg_press)

        self.btn_to_pos_press = Qt.QPushButton()
        self.prepare_button(self.btn_to_pos_press)
        self.btn_to_pos_press.setToolTip(
            'Moves the motor (while pressed) towards the Positive Software Limit')
        self.btn_to_pos_press.setIcon(
            Qt.QIcon("actions:media_seek_forward.svg"))
        btns_layout.addWidget(self.btn_to_pos_press)

        self.btn_to_pos = Qt.QPushButton()
        self.btn_to_pos.setToolTip(
            'Moves the motor towards the Positive Software Limit')
        self.prepare_button(self.btn_to_pos)
        self.btn_to_pos.setIcon(Qt.QIcon("actions:media_skip_forward.svg"))
        btns_layout.addWidget(self.btn_to_pos)

        btns_layout.addItem(Qt.QSpacerItem(
            1, 1, Qt.QSizePolicy.Expanding, Qt.QSizePolicy.Minimum))

        self.layout().addLayout(btns_layout, 1, 0, 1, 3)

        self.btn_step_down.clicked.connect(self.stepDown)
        self.btn_step_up.clicked.connect(self.stepUp)
        # self.btn_stop.clicked.connect(self.abort)
        self.btn_to_neg.clicked.connect(self.goNegative)
        self.btn_to_neg_press.pressed.connect(self.goNegative)
        self.btn_to_neg_press.released.connect(self.abort)
        self.btn_to_pos.clicked.connect(self.goPositive)
        self.btn_to_pos_press.pressed.connect(self.goPositive)
        self.btn_to_pos_press.released.connect(self.abort)

        # Align everything on top
        self.layout().addItem(Qt.QSpacerItem(
            1, 1, Qt.QSizePolicy.Minimum, Qt.QSizePolicy.Expanding), 2, 0, 1, 3)

        # IN ORDER TO BEHAVE AS EXPECTED REGARDING THE 'COMPACT VIEW' FEATURE
        # WE NEED TO SET THE 'EXPERTVIEW' WITHOUT ACCESSING THE taurusValueBuddy WHICH IS STILL NOT LINKED
        # SO WE ASSUME 'expertview is FALSE' AND WE HAVE TO AVOID self.setExpertView :-(
        # WOULD BE NICE THAT THE taurusValueBuddy COULD EMIT THE PROPER
        # SIGNAL...
        self.btn_to_neg.setVisible(False)
        self.btn_to_neg_press.setVisible(False)
        self.btn_to_pos.setVisible(False)
        self.btn_to_pos_press.setVisible(False)

        # IN EXPERT VIEW, WE HAVE TO FORWARD THE ''editingFinished()' SIGNAL
        # FROM TaurusValueLineEdit TO Switcher
        self.le_write_absolute.applied.connect(self.emitEditingFinished)
        self.btn_step_down.clicked.connect(self.emitEditingFinished)
        self.btn_step_up.clicked.connect(self.emitEditingFinished)
        self.btn_to_neg.clicked.connect(self.emitEditingFinished)
        self.btn_to_pos.clicked.connect(self.emitEditingFinished)

        # list of widgets used for edition
        editingWidgets = (self.le_write_absolute, self.cbAbsoluteRelative,
                          self.cb_step, self.btn_step_down,
                          self.btn_step_up, self.btn_to_neg,
                          self.btn_to_pos, self.btn_to_neg_press,
                          self.btn_to_pos_press)

        for w in editingWidgets:
            w.installEventFilter(self)

    def eventFilter(self, obj, event):
        '''reimplemented to intercept events from the subwidgets'''
        try:
            if obj in (self.btn_to_neg_press, self.btn_to_pos_press):
                if event.type() == Qt.QEvent.MouseButtonRelease:
                    self.emitEditingFinished()
        except AttributeError:
            # self.btn_to_neg_press, self.btn_to_pos_press may not exist now
            pass
        # emit editingFinished when focus out to a non-editing widget
        if event.type() == Qt.QEvent.FocusOut:
            focused = Qt.qApp.focusWidget()
            focusInChild = focused in self.findChildren(focused.__class__)
            if not focusInChild:
                self.emitEditingFinished()
        return False

    def cbAbsoluteRelativeChanged(self, abs_rel_option):
        abs_visible = abs_rel_option == 'Abs'
        rel_visible = abs_rel_option == 'Rel'
        if rel_visible:
            self.resetPendingOperations()
        self.le_write_absolute.setVisible(abs_visible)
        self.qw_write_relative.setVisible(rel_visible)

    @Qt.pyqtSlot()
    def stepDown(self):
        self.goRelative(-1)

    @Qt.pyqtSlot()
    def stepUp(self):
        self.goRelative(+1)

    @ProtectTaurusMessageBox(msg='An error occurred trying to move the motor.')
    def goRelative(self, direction):
        motor_dev = self.taurusValueBuddy().motor_dev
        if motor_dev is not None:
            increment = direction * float(self.cb_step.currentText())
            position = float(
                motor_dev.getAttribute('Position').read().rvalue.magnitude)
            target_position = position + increment
            motor_dev.getAttribute('Position').write(target_position)

    @Qt.pyqtSlot()
    @ProtectTaurusMessageBox(msg='An error occurred trying to move the motor.')
    def goNegative(self):
        motor_dev = self.taurusValueBuddy().motor_dev
        if motor_dev is not None:
            min_value = float(motor_dev.getAttribute('Position').min_value)
            motor_dev.getAttribute('Position').write(min_value)

    @Qt.pyqtSlot()
    @ProtectTaurusMessageBox(msg='An error occurred trying to move the motor.')
    def goPositive(self):
        motor_dev = self.taurusValueBuddy().motor_dev
        if motor_dev is not None:
            max_value = float(motor_dev.getAttribute('Position').max_value)
            motor_dev.getAttribute('Position').write(max_value)

    @Qt.pyqtSlot()
    @ProtectTaurusMessageBox(msg='An error occurred trying to abort the motion.')
    def abort(self):
        motor_dev = self.taurusValueBuddy().motor_dev
        if motor_dev is not None:
            motor_dev.abort()

    def prepare_button(self, btn):
        btn_policy = Qt.QSizePolicy(Qt.QSizePolicy.Fixed, Qt.QSizePolicy.Fixed)
        btn_policy.setHorizontalStretch(0)
        btn_policy.setVerticalStretch(0)
        btn.setSizePolicy(btn_policy)
        btn.setMinimumSize(25, 25)
        btn.setMaximumSize(25, 25)
        btn.setText('')

    def setExpertView(self, expertView):
        self.btn_to_neg.setVisible(expertView)
        self.btn_to_neg_press.setVisible(expertView)

        self.btn_to_pos.setVisible(expertView)
        self.btn_to_pos_press.setVisible(expertView)

        if expertView and self.taurusValueBuddy().motor_dev is not None:
            neg_sw_limit_enabled = self.taurusValueBuddy().motor_dev.getAttribute(
                'Position').min_value.lower() != 'not specified'
            self.btn_to_neg.setEnabled(neg_sw_limit_enabled)
            self.btn_to_neg_press.setEnabled(neg_sw_limit_enabled)

            pos_sw_limit_enabled = self.taurusValueBuddy().motor_dev.getAttribute(
                'Position').max_value.lower() != 'not specified'
            self.btn_to_pos.setEnabled(pos_sw_limit_enabled)
            self.btn_to_pos_press.setEnabled(pos_sw_limit_enabled)

    def setModel(self, model):
        if hasattr(self, 'taurusValueBuddy'):
            try:
                self.taurusValueBuddy().expertViewChanged.disconnect(
                    self.setExpertView)
            except TypeError:
                pass
        if model in (None, ''):
            TaurusWidget.setModel(self, model)
            self.le_write_absolute.setModel(model)
            return
        TaurusWidget.setModel(self, model + '/Position')
        self.le_write_absolute.setModel(model + '/Position#wvalue.magnitude')

        # Handle User/Expert View
        self.setExpertView(self.taurusValueBuddy()._expertView)
        self.taurusValueBuddy().expertViewChanged.connect(
            self.setExpertView)

    def keyPressEvent(self, key_event):
        if key_event.key() == Qt.Qt.Key_Escape:
            self.abort()
            key_event.accept()
        TaurusWidget.keyPressEvent(self, key_event)

    @Qt.pyqtSlot()
    def emitEditingFinished(self):
        self.applied.emit()

    def _config_limit(self, limit, hw_lim, sw_lim):
        """Configure sub-widgets according to limit information

        :param limit: which limit (negative or positive)
        :type limit: `Limit`
        :param hw_lim: hardware limit information
        :type hw_lim: `LimitInfo`
        :param hw_lim: software limit information
        :type hw_lim: `LimitInfo`
        """
        if limit == Limit.Negative:
            step_widget = self.btn_step_down
            to_widget = self.btn_to_neg
            to_press_widget = self.btn_to_neg_press
        elif limit == Limit.Positive:
            step_widget = self.btn_step_up
            to_widget = self.btn_to_pos
            to_press_widget = self.btn_to_pos_press
        else:
            raise ValueError("{} wrong limit".format(limit))
        active = hw_lim.active or sw_lim.active
        style_sheet = ""
        if active:
            style_sheet = 'QPushButton{{{}}}'.format(
                DEVICE_STATE_PALETTE.qtStyleSheet(PyTango.DevState.ALARM))
        step_widget.setEnabled(not active)
        allow_to = not active and sw_lim.exist
        to_widget.setEnabled(allow_to)
        to_press_widget.setEnabled(allow_to)
        step_widget.setStyleSheet(style_sheet)


##################################################
#                  UNITS WIDGET                  #
##################################################
class PoolMotorTVUnitsWidget(DefaultUnitsWidget):

    layoutAlignment = Qt.Qt.AlignTop

    def __init__(self, parent=None, designMode=False):
        DefaultUnitsWidget.__init__(self, parent, designMode)

    def setModel(self, model):
        if model in (None, ''):
            DefaultUnitsWidget.setModel(self, model)
            return
        DefaultUnitsWidget.setModel(self, model + '/Position')

##################################################
#                TV MOTOR WIDGET                 #
##################################################


class PoolMotorTV(TaurusValue):
    ''' A widget that displays and controls a pool Motor device.  It
    behaves as a TaurusValue.
    @TODO the view mode should be stored in the configuration
    @TODO the motor list should be stored in the configuration
    @TODO the selected radiobuttons (dial/indx/enc) and (abs/rel) should be stored in configuration
    @TODO it would be nice if the neg/pos limits could react also when software limits are 'active'
    @TODO expert view for read widget should include signals (indexer/encoder/inpos)...
    '''

    expertViewChanged = Qt.pyqtSignal(bool)

    def __init__(self, parent=None, designMode=False):
        TaurusValue.__init__(self, parent=parent, designMode=designMode)
        self.setLabelWidgetClass(PoolMotorTVLabelWidget)
        self.setReadWidgetClass(PoolMotorTVReadWidget)
        self.setWriteWidgetClass(PoolMotorTVWriteWidget)
        self.setUnitsWidgetClass(PoolMotorTVUnitsWidget)
        self.motor_dev = None
        self._expertView = False
        self.registerConfigProperty(
            self.getExpertView, self.setExpertView, 'ExpertView')
        self.limits_listener = None
        self.poweron_listener = None
        self.status_listener = None
        self.position_listener = None
        self.setExpertView(False)

    def setExpertView(self, expertView):
        self._expertView = expertView
        self.expertViewChanged.emit(expertView)

    def getExpertView(self):
        return self._expertView

    def minimumHeight(self):
        return None  # @todo: UGLY HACK to avoid subwidgets being forced to minimumheight=20

    def setModel(self, model):
        TaurusValue.setModel(self, model)

        # disconnect signals
        try:
            if self.limits_listener is not None:
                self.limits_listener.eventReceivedSignal.disconnect(
                    self.updateLimits)
        except TypeError:
            pass

        try:
            if self.poweron_listener is not None:
                self.poweron_listener.eventReceivedSignal.disconnect(
                    self.updatePowerOn)
        except TypeError:
            pass

        try:
            if self.status_listener is not None:
                self.status_listener.eventReceivedSignal.disconnect(
                    self.updateStatus)
        except TypeError:
            pass

        try:
            if self.position_listener is not None:
                self.position_listener.eventReceivedSignal.disconnect(
                    self.updatePosition)
        except TypeError:
            pass

        try:
            # remove listeners
            if self.motor_dev is not None:
                if self.hasHwLimits():
                    self.motor_dev.getAttribute(
                        'Limit_Switches').removeListener(self.limits_listener)
                if self.hasPowerOn():
                    self.motor_dev.getAttribute(
                        'PowerOn').removeListener(self.poweron_listener)
                self.motor_dev.getAttribute(
                    'Status').removeListener(self.status_listener)
                self.motor_dev.getAttribute(
                    'Position').removeListener(self.position_listener)

            if model == '' or model is None:
                self.motor_dev = None
                return
            self.motor_dev = taurus.Device(model)
            # CONFIGURE A LISTENER IN ORDER TO UPDATE LIMIT SWITCHES STATES
            self.limits_listener = TaurusAttributeListener()
            if self.hasHwLimits():
                self.limits_listener.eventReceivedSignal.connect(
                    self.updateLimits)
                self._limit_switches = self.motor_dev.getAttribute(
                    'Limit_Switches')
                self._limit_switches.addListener(self.limits_listener)

            # CONFIGURE AN EVENT RECEIVER IN ORDER TO PROVIDE POWERON <-
            # True/False EXPERT OPERATION
            self.poweron_listener = TaurusAttributeListener()
            if self.hasPowerOn():
                self.poweron_listener.eventReceivedSignal.connect(
                    self.updatePowerOn)
                self._poweron = self.motor_dev.getAttribute('PowerOn')
                self._poweron.addListener(self.poweron_listener)

            # CONFIGURE AN EVENT RECEIVER IN ORDER TO UPDATED STATUS TOOLTIP
            self.status_listener = TaurusAttributeListener()
            self.status_listener.eventReceivedSignal.connect(
                self.updateStatus)
            self._status = self.motor_dev.getAttribute('Status')
            self._status.addListener(self.status_listener)

            # CONFIGURE AN EVENT RECEIVER IN ORDER TO ACTIVATE LIMIT BUTTONS ON
            # SOFTWARE LIMITS
            self.position_listener = TaurusAttributeListener()
            self.position_listener.eventReceivedSignal.connect(
                self.updatePosition)
            self._position = self.motor_dev.getAttribute('Position')
            self._position.addListener(self.position_listener)
            self.motor_dev.getAttribute('Position').enablePolling(force=True)

            self.setExpertView(self._expertView)
        except Exception as e:
            self.warning("Exception caught while setting model: %s", repr(e))
            self.motor_dev = None
            return

    def hasPowerOn(self):
        try:
            return hasattr(self.motor_dev, 'PowerOn')
        except:
            return False

    def hasHwLimits(self):
        try:
            return hasattr(self.motor_dev, 'Limit_Switches')
        except:
            return False

    def updateLimits(self, limits, position=None):
        if isinstance(limits, dict):
            limits = limits["limits"]

        pos_hw_lim = eval_hw_limit(Limit.Positive, limits[Limit.Positive])
        neg_hw_lim = eval_hw_limit(Limit.Negative, limits[Limit.Negative])
        position_attr = self.motor_dev.getAttribute('Position')
        pos_sw_lim = eval_sw_limit(Limit.Positive, position_attr)
        neg_sw_lim = eval_sw_limit(Limit.Negative, position_attr)

        read_widget = self.readWidget(followCompact=True)
        read_widget._config_limit(Limit.Positive, pos_hw_lim, pos_sw_lim)
        read_widget._config_limit(Limit.Negative, neg_hw_lim, neg_sw_lim)

        write_widget = self.writeWidget(followCompact=True)
        write_widget._config_limit(Limit.Positive, pos_hw_lim, pos_sw_lim)
        write_widget._config_limit(Limit.Negative, neg_hw_lim, neg_sw_lim)

    def updatePowerOn(self, poweron='__no_argument__'):
        if poweron == '__no_argument__':
            msg = 'updatePowerOn called without args (bug in old PyQt). Ignored'
            self.debug(msg)
            return
        btn_text = 'Set ON'
        if poweron:
            btn_text = 'Set OFF'
        self.labelWidget().btn_poweron.setText(btn_text)

    def updateStatus(self, status):
        # SHOULD THERE BE A BETTER METHOD FOR THIS UPDATE?
        # IF THIS IS NOT DONE, THE TOOLTIP IS NOT CALCULATED EVERY TIME
        # TaurusLabel.updateStyle DIDN'T WORK, SO I HAD TO GO DEEPER TO THE CONTROLLER...
        # self.labelWidget().lbl_alias.updateStyle()
        self.labelWidget().lbl_alias.controllerUpdate()

    def updatePosition(self, position='__no_argument__'):
        if position == '__no_argument__':
            msg = 'updatePowerOn called without args (bug in old PyQt). Ignored'
            self.debug(msg)
            return
        # we do not need the position for nothing...
        # we just want to check if any software limit is 'active'
        # and updateLimits takes care of it
        if self.motor_dev is not None:
            limit_switches = [None, None, None]
            if self.hasHwLimits():
                limit_switches = self.motor_dev.getAttribute(
                    'Limit_switches').read().rvalue
                # print "update limits", limit_switches
            self.updateLimits(limit_switches, position=position)

    def hasEncoder(self):
        try:
            return hasattr(self.motor_dev, 'Encoder')
        except:
            return False

    def showTangoAttributes(self):
        model = self.getModel()
        taurus_attr_form = TaurusAttrForm()
        taurus_attr_form.setMinimumSize(Qt.QSize(555, 800))
        taurus_attr_form.setModel(model)
        taurus_attr_form.setWindowTitle(
            '%s Tango Attributes' % taurus.Factory().getDevice(model).getSimpleName())
        taurus_attr_form.show()

    # def showEvent(self, event):
    ###     TaurusValue.showEvent(self, event)
    # if self.motor_dev is not None:
    # self.motor_dev.getAttribute('Position').enablePolling(force=True)
    ###
    # def hideEvent(self, event):
    ###     TaurusValue.hideEvent(self, event)
    # if self.motor_dev is not None:
    # self.motor_dev.getAttribute('Position').disablePolling()

###################################################
# A SIMPLER WIDGET THAT MAY BE USED OUTSIDE FORMS #
###################################################


class PoolMotor(TaurusFrame):
    ''' A widget that displays and controls a pool Motor device.
    '''

    def __init__(self, parent=None, designMode=False):
        TaurusFrame.__init__(self, parent, designMode)

        self.setLayout(Qt.QGridLayout())
        self.layout().setContentsMargins(0, 0, 0, 0)
        self.layout().setSpacing(0)

        self.setFrameShape(Qt.QFrame.Box)

        self.pool_motor_tv = PoolMotorTV(self)

    def setModel(self, model):
        self.pool_motor_tv.setModel(model)
        try:
            self.motor_dev = taurus.Device(model)
        except:
            return


def main():

    import sys
    import taurus.qt.qtgui.application
    import taurus.core.util.argparse
    from taurus.qt.qtgui.panel import TaurusForm
    parser = taurus.core.util.argparse.get_taurus_parser()
    parser.usage = "%prog [options] [<motor1> [<motor2>] ...]"

    app = taurus.qt.qtgui.application.TaurusApplication(cmd_line_parser=parser)
    args = app.get_command_line_args()

    #models = ['tango://controls02:10000/motor/gcipap10ctrl/8']
    models = ['motor/motctrl13/3']

    if len(args) > 0:
        models = args

    w = Qt.QWidget()
    w.setLayout(Qt.QVBoxLayout())

    tests = []
    # tests.append(1)
    tests.append(2)
    # tests.append(3)
    # tests.append(4)

    # 2) Test PoolMotorTV motor widget
    form_tv = TaurusForm()
    form_tv.setModifiableByUser(True)
    tv_widget_class = 'sardana.taurus.qt.qtgui.extra_pool.PoolMotorTV'
    tv_tgclass_map = {'SimuMotor': (tv_widget_class, (), {}),
                      'Motor': (tv_widget_class, (), {}),
                      'PseudoMotor': (tv_widget_class, (), {})}
    form_tv.setCustomWidgetMap(tv_tgclass_map)

    if 2 in tests:
        form_tv.setModel(models)
        w.layout().addWidget(form_tv)
        form_tv.setCompact(True)

    # 3) Test Stand-Alone PoolMotor widget
    # New approach would be to let PoolMotorTV live outside a TaurusForm.... but inside a GridLayout
    # Carlos already said this is not a good approach but...
    if 3 in tests:
        for motor in models:
            motor_widget = PoolMotor()
            motor_widget.setModel(motor)
            w.layout().addWidget(motor_widget)

    w.show()

    sys.exit(app.exec_())

if __name__ == '__main__':
    main()



##############################################################################
##
# This file is part of Sardana
##
# http://www.sardana-controls.org/
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""This module contains macros that demonstrate the usage of motion"""

__all__ = ["move_async"]

__docformat__ = 'restructuredtext'

from sardana.macroserver.macro import *


class move_async(Macro):
    """A macro that executes an asynchronous movement of a motor. The movement
    can be cancelled by a Ctrl-C signal.

    This macro is part of the examples package. It was written for
    demonstration purposes"""

    param_def = [['moveable', Type.Moveable, None, 'moveable to be moved'],
                 ['pos', Type.Float, None, 'target position'],
                 ]

    def run(self, moveable, pos):
        motion = self.getMotion([moveable])
        self.info('initial position: %s' % motion.readPosition())
        _id = motion.startMove([pos])
        try:
            # Do whatever here (while the moveable is moving)
            self.info('state: %s' % motion.readState())
            # End Do whatever here
        finally:
            motion.waitMove(id=_id)
            # this line will not be printed in case of abort (Ctrl-C)
            # due to https://sourceforge.net/p/sardana/tickets/9/
            self.info('final position: %s' % motion.readPosition())



#!/usr/bin/env python

##############################################################################
##
# This file is part of Sardana
##
# http://www.sardana-controls.org/
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""This module is part Sardana Python library. It defines the base clases
for Sardana buffers"""



__all__ = ["SardanaBuffer", "LateValueException", "EarlyValueException"]

import weakref

from collections import OrderedDict

from .sardanavalue import SardanaValue
from .sardanaevent import EventGenerator, EventType
from .sardanaexception import SardanaException


class LateValueException(SardanaException):
    """Exception indicating that a given value is not present in the buffer and
    will not arrive any more (newer value(s) were already added to the buffer).
    """
    pass


class EarlyValueException(SardanaException):
    """Exception indicating that a given value is not present in the buffer but
    there is still a chance that it will arrive (no newer values were added to
    the buffer yet).
    """
    pass


class SardanaBuffer(EventGenerator):
    """Buffer for SardanaValue objects. Each value is identified by an unique
    idx and all values are organized based on the order of addition to the
    buffer

    ..todo:: Eliminate the last_chunk - it is not really necessary and just
     consumes memory.
    ..todo:: It is better to fire events in form of a list with tuples
    of idx and value objects.
    """

    def __init__(self, obj=None, name=None, persistent=False, **kwargs):
        """Construct SardanaBuffer object

        :param obj: the object which owns this buffer
        :type obj: obj
        :param name: object name
        :type name: :obj:`str`
        :param persistent: whether values are kept in the buffer until
            being explicitly removed (True) or just until firing the next event
            (False)
        :type persistent: bool
        """
        super(SardanaBuffer, self).__init__(**kwargs)
        if obj is not None:
            obj = weakref.ref(obj)
        self._obj = obj
        self.name = name or self.__class__.__name__
        self._persistent = persistent
        self._buffer = OrderedDict()
        self._next_idx = 0
        self._last_chunk = None

    def __len__(self):
        return self._buffer.__len__()

    def get_obj(self):
        """Returns the object which owns this buffer

        :return: the object which owns this buffer
        :rtype: obj"""
        return self._get_obj()

    def _get_obj(self):
        obj = self._obj
        if obj is not None:
            obj = obj()
        return obj

    def get_value(self, idx):
        """Return value of a given index.

        :param idx: index of the value to be returned
        :type idx: int
        :return: the value corresponding to the idx
        :rtype: object
        """
        return self.get_value_obj(idx).value

    def get_value_obj(self, idx):
        """Return the value object of a given index.

        :param idx: index of the value to be returned
        :type idx: int
        :return: the value object corresponding to the idx
        :rtype: SardanaValue
        """
        try:
            return self._buffer[idx]
        except KeyError:
            msg = "value with %s index is not in buffer" % idx
            if self.next_idx > idx:
                raise LateValueException(msg)
            else:
                raise EarlyValueException(msg)

    def append(self, value, idx=None):
        """Append a single value at the end of the buffer with a given index.

        :param value: value to be appended to the buffer
        :type param: SardanaValue or any object
        :param idx: at which index append the value, None means append at the
            end of the buffer
        :type idx: int
        :param persistent: whether value should be added to a persistent
            buffer or just as a last chunk
        :type param: bool
        """
        if idx is None:
            idx = self._next_idx
        self._last_chunk = OrderedDict()
        if not isinstance(value, SardanaValue):
            value = SardanaValue(value)
        self._last_chunk[idx] = value
        if self._persistent:
            self._buffer[idx] = value
        self._next_idx = idx + 1
        self.fire_add_event()

    def extend(self, values, initial_idx=None):
        """Extend buffer with a list of objects assigning them consecutive
        indexes.

        :param values: objects that extend the buffer
        :type values: list<object>
        :param initial_idx: at which index append the first object,
            the rest of them will be assigned the next consecutive indexes,
            None means assign at the end of the buffer
        :type initial_idx: int
        """
        if initial_idx is None:
            initial_idx = self._next_idx
        self._last_chunk = OrderedDict()
        for idx, value in enumerate(values, initial_idx):
            if not isinstance(value, SardanaValue):
                value = SardanaValue(value)
            self._last_chunk[idx] = value
            if self._persistent:
                self._buffer[idx] = value
        self._next_idx = idx + 1
        self.fire_add_event()

    def remove(self, idx):
        """Remove value object of a given index.

        :param idx: index of the value to be returned
        :type idx: int
        :return: the value object corresponding to the idx
        :rtype: object
        """
        try:
            return self._buffer.pop(idx)
        except KeyError:
            msg = "value with %s index is not in buffer" % idx
            raise KeyError(msg)

    def fire_add_event(self, propagate=1):
        """Fires an event to the listeners of the object which owns this
        buffer.

        :param propagate:
            0 for not propagating, 1 to propagate, 2 propagate with priority
        :type propagate: int
        """
        evt_type = EventType(self.name, priority=propagate)
        self.fire_event(evt_type, self.last_chunk)

    def clear(self):
        self._next_idx = 0
        self._buffer = OrderedDict()

    def get_last_chunk(self):
        return self._last_chunk

    def get_next_idx(self):
        return self._next_idx

    def get_persistent(self):
        return self._persistent

    def set_persistent(self, persistent):
        self._persistent = persistent

    obj = property(get_obj, "container object for this buffer")
    persistent = property(get_persistent, set_persistent, "flag whether this "
                          "buffer stores objects persistently")
    last_chunk = property(get_last_chunk,
                          doc="chunk with last value(s) added to this buffer")
    next_idx = property(get_next_idx,
                        doc="index that will be automatically assigned to the "
                            "next value added to this buffer (if not "
                            "explicitly assigned by the user)")



#!/usr/bin/env python

##############################################################################
##
# This file is part of Sardana
##
# http://www.sardana-controls.org/
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""This module contains a definition for ViewOptions"""




__all__ = ['ViewOption']


def ViewOptionMeta(name, bases, attrs):
    for k in attrs['_DEFAULT_VIEW_OPTIONS']:
        attrs[k] = k
    return type(name, bases, attrs)


class ViewOption(object, metaclass=ViewOptionMeta):
    _DEFAULT_VIEW_OPTIONS = {
        'ShowDial': True,
        'ShowCtrlAxis': False,
        'PosFormat': -1,
        'OutputBlock': False,
        'DescriptionLength': 60
    }

    @classmethod
    def init_options(cls, d):
        '''Init missing options with the default values.
        '''
        diff = set(cls._DEFAULT_VIEW_OPTIONS).difference(d)
        if len(diff) > 0:
            for key in diff:
                d[key] = cls._DEFAULT_VIEW_OPTIONS[key]
        return d

    @classmethod
    def reset_option(cls, d, name):
        if name in cls._DEFAULT_VIEW_OPTIONS:
            d[name] = cls._DEFAULT_VIEW_OPTIONS[name]
        else:
            del d[name]



#!/usr/bin/env python

##############################################################################
##
# This file is part of Sardana
##
# http://www.sardana-controls.org/
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""
This module provides TaurusValue item factories to be registered as providers
of custom TaurusForm item widgets.
"""

from . import PoolMotorTV, PoolChannelTV, PoolIORegisterTV, _PoolChannelTV

T_FORM_POOL_WIDGET_MAP = {
    "SimuMotor": PoolMotorTV,
    "Motor": PoolMotorTV,
    "PseudoMotor": PoolMotorTV,
    "PseudoCounter": _PoolChannelTV,
    "CTExpChannel": PoolChannelTV,
    "ZeroDExpChannel": _PoolChannelTV,
    "OneDExpChannel": PoolChannelTV,
    "TwoDExpChannel": PoolChannelTV,
    "IORegister": PoolIORegisterTV,
}


def pool_item_factory(model):
    """
    Taurus Value Factory to be registered as a TaurusForm item factory plugin

    :param model: taurus model object

    :return: custom TaurusValue class
    """
    # TODO: use sardana element types instead of tango classes
    try:
        key = model.getDeviceProxy().info().dev_class
        klass = T_FORM_POOL_WIDGET_MAP[key]
    except Exception:
        return None
    return klass()



#!/usr/bin/env python

##############################################################################
##
# This file is part of Sardana
##
# http://www.sardana-controls.org/
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""This module contains the class definition for the MacroServer environment
manager"""

__all__ = ["MacroServerException", "MacroServerExceptionList", "MissingEnv",
           "UnknownEnv", "UnknownMacro", "UnknownMacroLibrary",
           "UnknownRecorder", "MacroWrongParameterType", "LibraryError",
           "InterruptException", "StopException", "AbortException",
           "ReleaseException", "InputCancelled"]

__docformat__ = 'restructuredtext'

from sardana.taurus.core.tango.sardana.pool import InterruptException, \
    StopException, AbortException, ReleaseException

from sardana.sardanaexception import SardanaException, SardanaExceptionList, \
    UnknownCode, UnknownLibrary, LibraryError


class MacroServerException(SardanaException):
    pass


class MacroServerExceptionList(SardanaExceptionList):
    pass


class MissingEnv(MacroServerException):
    pass


class UnknownEnv(MacroServerException):
    pass


class UnknownMacro(UnknownCode):
    pass


class UnknownRecorder(UnknownCode):
    pass


class UnknownMacroLibrary(UnknownLibrary):
    pass


class MacroWrongParameterType(MacroServerException):
    pass


class InputCancelled(MacroServerException):
    pass



#!/usr/bin/env python

##############################################################################
##
# This file is part of Sardana
##
# http://www.sardana-controls.org/
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""This module contains a taurus text editor widget."""

__all__ = ["SardanaEditor"]

__docformat__ = 'restructuredtext'

import os
import os.path as osp
import tempfile
import shutil
import functools

from taurus.external.qt import Qt

from taurus.qt.qtgui.base import TaurusBaseWidget
from taurus.qt.qtgui.editor import TaurusBaseEditor
from taurus.qt.qtgui.dialog import ProtectTaurusMessageBox

# consider adding an utility in taurusedirot to create actions
from spyder.utils.qthelpers import create_action

from .macrotree import MacroSelectionDialog
from .elementtree import SardanaElementTreeWidget

from sardana.taurus.qt.qtcore.tango.sardana.model import SardanaBaseProxyModel, \
    SardanaElementTypeModel, SardanaTypeTreeItem, SardanaRootTreeItem

from .sardanabasewizard import SardanaBaseWizard, SardanaBasePage

_MACRO_LIB_TEMPLATE = """#!/usr/bin/env python
{copyright}
# don't forget to place every new macro here!
__all__ = []

__docformat__ = 'restructuredtext'

{non_sardana_imports}
from sardana.macroserver.macro import macro, Macro, Type
{sardana_imports}
# Place your code here!

"""

_MACRO_CLASS_TEMPLATE = """

class {macro_name}(Macro):
    \"\"\"{macro_name} description.\"\"\"

    # uncomment the following lines as necessary. Otherwise you may delete them
    #param_def = []
    #result_def = []
    #hints = {}
    #env = (,)

    # uncomment the following lines if need prepare. Otherwise you may delete them
    #def prepare(self):
    #    pass

    def run(self):
        pass

"""

_MACRO_FUNCTION_TEMPLATE = """

@macro()
def {macro_name}(self):
    self.output("Running {macro_name}...")

"""


class NewElementWizard(SardanaBaseWizard):
    pass


class ChooseElementTypePage(SardanaBasePage):

    def __init__(self, parent=None):
        SardanaBasePage.__init__(self, parent)

        self.setTitle('Please select type of element you wish to create')
        layout = Qt.QVBoxLayout()
        self.setLayout(layout)


class SardanaLibProxyModel(SardanaBaseProxyModel):

    ALLOWED_TYPES = 'MacroLibrary',  # 'ControllerLibrary',

    def filterAcceptsRow(self, sourceRow, sourceParent):
        sourceModel = self.sourceModel()
        idx = sourceModel.index(sourceRow, 0, sourceParent)
        treeItem = idx.internalPointer()

        if isinstance(treeItem, SardanaRootTreeItem):
            return True
        if isinstance(treeItem, SardanaTypeTreeItem):
            return treeItem.itemData() in self.ALLOWED_TYPES
        return True


class SardanaLibTreeWidget(SardanaElementTreeWidget):

    KnownPerspectives = {"Type": {
        "label": "By lib",
        "icon": ":/python-package.png",
        "tooltip": "View elements by library",
        "model": [SardanaLibProxyModel, SardanaElementTypeModel],
    },
    }
    DftPerspective = "Type"


class SardanaEditor(TaurusBaseEditor, TaurusBaseWidget):

    def __init__(self, parent=None, designMode=None):
        name = self.__class__.__name__
        self._base_tmp_dir = None
        self._tmp_dir = None
        self._is_filesystem_prepared = False
        self.call__init__wo_kw(TaurusBaseEditor, parent)
        self.call__init__(TaurusBaseWidget, name, designMode=designMode)
        self._elementTree = elementTree = \
            SardanaLibTreeWidget(self, with_navigation_bar=False,
                                 with_filter_widget=False,)
        elementTree.treeView().setColumnHidden(1, True)
        self._elementTree.itemDoubleClicked.connect(
            self.on_element_clicked)
        self.insertWidget(0, self._elementTree)
        self.setAutoTooltip(False)

    def setTemporaryDirectory(self, tmp_dir):
        self._base_tmp_dir = tmp_dir
        self._is_filesystem_prepared = False

    def getTemporaryDirectory(self):
        return self._base_tmp_dir

    def createMenuActions(self):
        on_save = functools.partial(self.on_save, apply=False)
        on_save_apply = functools.partial(self.on_save, apply=True)

        self.new_action = create_action(self, "New...",
                                        icon=Qt.QIcon.fromTheme(
                                            'document-new'),
                                        tip="Create a new macro or controller class",
                                        triggered=self.on_new, shortcut=Qt.QKeySequence.New)

        self.open_action = create_action(self, "Open...",
                                         icon=Qt.QIcon.fromTheme(
                                             'document-open'),
                                         tip="Open macro(s) or controller(s)",
                                         triggered=self.on_open, shortcut=Qt.QKeySequence.Open)
        self.save_action = create_action(self, "Save",
                                         icon=Qt.QIcon.fromTheme(
                                             'document-save'),
                                         tip="Save the current selected item",
                                         triggered=on_save)
        self.save_and_apply_action = create_action(self, "Save && apply",
                                                   triggered=on_save_apply,
                                                   icon=Qt.QIcon.fromTheme(
                                                       'document-save'),
                                                   tip="Save the current selected item and apply the new code",
                                                   shortcut=Qt.QKeySequence.Save)
        self.revert_action = create_action(self, "Revert",
                                           icon=Qt.QIcon.fromTheme(
                                               'edit-undo'),
                                           tip="Revert the current selected item code",
                                           triggered=self.on_revert)

        io_actions = [self.new_action, self.open_action, self.save_action,
                      self.revert_action]
        actions = [self.new_action, self.open_action, self.save_action,
                   self.save_and_apply_action, self.revert_action]
        return actions, io_actions

    def register_editorstack(self, editorstack):
        TaurusBaseEditor.register_editorstack(self, editorstack)
        editorstack.refresh_save_all_action.connect(
            self.refresh_save_and_apply_action)

    def refresh_save_and_apply_action(self):
        self.save_and_apply_action.setEnabled(self.save_action.isEnabled())

    def on_element_clicked(self, item, item_column):
        item_data = item.itemData()
        if isinstance(item_data, str):
            return
        interfaces = item_data.interfaces
        if 'MacroCode' in interfaces:
            self.open_macros([item_data])
        elif 'MacroLibrary' in interfaces:
            self.open_macro_libraries([item_data])

    @ProtectTaurusMessageBox(title="A error occured trying to create a class")
    def on_new(self, checked):

        elem_types = "Macro function", "Macro class", "Macro library", \
            "Motor controller class", "Counter/Timer controller class", \
            "Pseudo motor controller class"

        msg = "Please select type of element you wish to create"
        elem_type, ok = Qt.QInputDialog.getItem(self, "New", msg, elem_types,
                                                current=0, editable=False)
        if not ok:
            return

        idx = elem_types.index(elem_type)
        if idx == 0:
            return self.new_macro_function()
        elif idx == 1:
            return self.new_macro_class()
        elif idx == 2:
            return self.new_macro_library()
        raise NotImplementedError("Sorry! Not implemented yet.")

    def new_macro(self, template):
        macro_server = self.getModelObj()

        msg = "Please select the library where you want to place the new macro"

        macro_libraries = macro_server.getElementsOfType("MacroLibrary")
        macro_lib_names = sorted(macro_libraries.keys())
        macro_lib_name, ok = Qt.QInputDialog.getItem(self, "Macro library", msg,
                                                     macro_lib_names, current=0, editable=False)
        if not ok:
            return
        macro_lib_name = str(macro_lib_name)
        macro_lib = macro_libraries[macro_lib_name]

        fname, path = macro_lib.file_path, macro_lib.path
        # transform into relative path
        fname = fname[fname.index(osp.sep) + 1:]
        local_filename = osp.join(self._tmp_dir, fname)

        msg = "Please give new macro name"

        macros = macro_lib.elements
        valid = False
        while not valid:
            macro_name, ok = Qt.QInputDialog.getText(self, "Macro name", msg)
            if not ok:
                return
            if macro_name in macros:
                res = Qt.QMessageBox.information(self, "Macro already exists",
                                                 "A macro named '%s' already exists in '%s'.\n"
                                                 "Please give a different macro name"
                                                 % (macro_name, macro_lib_name),
                                                 Qt.QMessageBox.Ok | Qt.QMessageBox.Cancel, Qt.QMessageBox.Ok)
                if res == Qt.QMessageBox.Cancel:
                    return
                continue
            valid = True

        pars = dict(macro_name=macro_name)
        code = template.format(**pars)

        editorstack = self.editorStack()
        idx = editorstack.has_filename(local_filename)
        if idx is None:
            if not self._prepare_path(path):
                raise Exception("Error trying to prepare path for %s", path)
            self.debug("Creating local file %s...", local_filename)
            fname, lib_code, line = macro_server.GetMacroCode(
                (macro_lib_name,))
            fd = open(local_filename, "w")
            fd.write(lib_code)
            fd.close()
            self.debug("Loading local file %s...", local_filename)
            self.load(local_filename, goto=None)
        else:
            pass
        editorstack.set_current_filename(local_filename)
        editor = editorstack.get_current_editor()
        editor.set_cursor_position(editor.get_position("eof"))
        editor.append(code)
        return macro_lib, macro_name

    def new_macro_class(self):
        macro_info = self.new_macro(_MACRO_CLASS_TEMPLATE)
        if macro_info is None:
            return
        macro_lib, macro_name = macro_info

    def new_macro_function(self):
        macro_info = self.new_macro(_MACRO_FUNCTION_TEMPLATE)
        if macro_info is None:
            return
        macro_lib, macro_name = macro_info

    def new_macro_library(self):
        ms = self.getModelObj()
        ms_path = ms.getMacroPathObj()
        directory, ok = Qt.QInputDialog.getItem(self, "New macro module",
                                                "Select the directory where you want the new macro module to "
                                                "be placed",
                                                ms_path.macro_path, current=0, editable=False)
        if not ok:
            return
        directory = str(directory)
        ok = 0
        while not ok:
            mod, ok = Qt.QInputDialog.getText(self, "New macro module",
                                              "Module name:", Qt.QLineEdit.Normal, "")
            if not ok:
                return
            mod = str(mod)
            m, ext = osp.splitext(mod)
            if len(ext):
                if ext != ".py":
                    Qt.QMessageBox.critical(self, "Wrong extension",
                                            "When given, file extension<b> MUST</b> be <code>.py</code>.")
                    ok = 0
                    continue
            else:
                mod = mod + ".py"
        self._prepare_path(directory)
        filename = osp.join(directory, mod)
        # transform into relative path
        rel_filename = filename[filename.index(osp.sep) + 1:]
        local_filename = osp.join(self._tmp_dir, rel_filename)
        f = open(local_filename, "w")

        # TODO: ask for additional imports
        # TODO: check if door environment has copyright variable

        pars = dict(copyright="", non_sardana_imports="",
                    sardana_imports="")
        code = _MACRO_LIB_TEMPLATE.format(**pars)
        f.write(code)
        f.close()
        self.editorStack().load(local_filename)

    @ProtectTaurusMessageBox(title="An error occured trying to open a macro class")
    def on_open(self, checked):
        ms = self.getModelObj()
        ms_tree = MacroSelectionDialog(self, model_name=ms.getNormalName())
        ms_tree.exec_()
        if ms_tree.result() != Qt.QDialog.Accepted:
            return
        self.open_macros(ms_tree.getSelectedMacros())

    @ProtectTaurusMessageBox(title="An error occured trying to open macro(s)")
    def open_macros(self, macros):
        editorstack = self.editorStack()

        all_any = Qt.QMessageBox.YesToAll, Qt.QMessageBox.NoToAll
        yes_any = Qt.QMessageBox.Yes, Qt.QMessageBox.YesToAll
        no_any = Qt.QMessageBox.No, Qt.QMessageBox.NoToAll

        last_answer = Qt.QMessageBox.No
        for macro_info in macros:
            name, fname = macro_info.name, macro_info.file_path
            module = macro_info.module
            # transform into relative path
            fname = fname[fname.index(osp.sep) + 1:]
            local_filename = osp.join(self._tmp_dir, fname)

            idx = editorstack.has_filename(local_filename)
            if idx is not None and last_answer not in all_any:
                last_answer = Qt.QMessageBox.question(self,
                                                      "Macro file '{0}' already opened".format(
                                                          module),
                                                      "All changes to <b>{0}</b> will be lost."
                                                      "<br>Do you want to revert file from the server ignoring "
                                                      "any possible changes you (may) have made?".format(
                                                          module),
                                                      Qt.QMessageBox.Yes | Qt.QMessageBox.YesToAll |
                                                      Qt.QMessageBox.No | Qt.QMessageBox.NoToAll,
                                                      Qt.QMessageBox.No)

            if idx is None or last_answer in yes_any:
                if not self._prepare_path(macro_info.path):
                    Qt.QMessageBox.warning(self,
                                           "Error trying to prepare '{0}'".format(
                                               module),
                                           "An error occured trying to prepare '{0}'".format(
                                               module),
                                           Qt.QMessageBox.Ok, Qt.QMessageBox.Ok)
                    continue
                _, code, line = self.get_macro_code(module, name)
                line = int(line)
                self.debug("Creating local file %s...", local_filename)
                fd = open(local_filename, "w")
                fd.write(code)
                fd.close()
                if idx is None:
                    self.debug("Loading local file %s...", local_filename)
                    self.load(local_filename, goto=line)
                else:
                    self.debug("Reloading local file %s...", local_filename)
                    self.reload(idx, filename=local_filename, goto=line)
            elif last_answer in no_any:
                self.set_current_filename(local_filename)

    @ProtectTaurusMessageBox(title="An error occured trying to open macro(s)")
    def open_macro_libraries(self, macro_libraries):
        editorstack = self.editorStack()

        all_any = Qt.QMessageBox.YesToAll, Qt.QMessageBox.NoToAll
        yes_any = Qt.QMessageBox.Yes, Qt.QMessageBox.YesToAll
        no_any = Qt.QMessageBox.No, Qt.QMessageBox.NoToAll

        last_answer = Qt.QMessageBox.No
        for macro_library in macro_libraries:
            name, fname = macro_library.name, macro_library.file_path
            module = macro_library.module
            # transform into relative path
            fname = fname[fname.index(osp.sep) + 1:]
            local_filename = osp.join(self._tmp_dir, fname)

            idx = editorstack.has_filename(local_filename)
            if idx is not None and last_answer not in all_any:
                last_answer = Qt.QMessageBox.question(self,
                                                      "Macro file '{0}' already opened".format(
                                                          module),
                                                      "All changes to <b>{0}</b> will be lost."
                                                      "<br>Do you want to revert file from the server ignoring "
                                                      "any possible changes you (may) have made?".format(
                                                          module),
                                                      Qt.QMessageBox.Yes | Qt.QMessageBox.YesToAll |
                                                      Qt.QMessageBox.No | Qt.QMessageBox.NoToAll,
                                                      Qt.QMessageBox.No)

            if idx is None or last_answer in yes_any:
                if not self._prepare_path(macro_library.path):
                    Qt.QMessageBox.warning(self,
                                           "Error trying to prepare '{0}'".format(
                                               module),
                                           "An error occured trying to prepare '{0}'".format(
                                               module),
                                           Qt.QMessageBox.Ok, Qt.QMessageBox.Ok)
                    continue
                _, code, line = self.get_macro_code(module)
                line = int(line)
                self.debug("Creating local file %s...", local_filename)
                fd = open(local_filename, "w")
                fd.write(code)
                fd.close()
                if idx is None:
                    self.debug("Loading local file %s...", local_filename)
                    self.load(local_filename, goto=line)
                else:
                    self.debug("Reloading local file %s...", local_filename)
                    self.reload(idx, filename=local_filename, goto=line)
            elif last_answer in no_any:
                self.set_current_filename(local_filename)

    @ProtectTaurusMessageBox(msg="A error occured trying to save")
    def on_save(self, checked, apply=True):
        editorstack = self.editorStack()
        # Save the currently edited file
        if not editorstack.get_stack_count():
            return
        index = editorstack.get_stack_index()
        res = editorstack.save(index=index)
        file_info = editorstack.data[index]
        if not res:
            return
        local_filename = file_info.filename
        fd = open(local_filename, "r")
        code = fd.read()
        fd.close()
        remote_filename = local_filename[len(self._tmp_dir):]
        self.set_macro_code(remote_filename, code, apply)

    def on_revert(self, checked):
        self.editorStack().revert()

    def reload_macro_lib(self, module_name):
        pass

    def set_macro_code(self, filename, code, apply=True):
        ms = self.getModelObj()
        if apply:
            apply = "true"
        else:
            apply = "false"
        return ms.SetMacroCode((filename, code, apply))

    def get_macro_code(self, module_name, macro_name=None):
        ms = self.getModelObj()
        pars = [module_name]
        if macro_name is not None:
            pars.append(macro_name)
        return ms.GetMacroCode(pars)

    def setModel(self, model_name):
        TaurusBaseWidget.setModel(self, model_name)
        self._elementTree.setModel(model_name)
        self._is_filesystem_prepared = False
        self.prepare_filesystem()

    def _prepare_path(self, p):
        if not self.prepare_filesystem():
            return False
        p = p[p.index(osp.sep) + 1:]  # transform into relative path
        p = osp.join(self._tmp_dir, p)
        if not osp.exists(p):
            os.makedirs(p)
        return True

    def prepare_filesystem(self):
        """Prepares a temporary directory to store the macro files locally"""
        if self._is_filesystem_prepared:
            return True
        ms = self.getModelObj()
        if ms is None:
            self.warning("Could not prepare local filesystem to store macros")
            return False
        ms_name = ms.getSimpleName().replace('/', '_')
        self._tmp_dir = tempfile.mkdtemp(
            prefix=ms_name, dir=self._base_tmp_dir)
        self._is_filesystem_prepared = True
        return True

    def closeEvent(self, event):
        if self._is_filesystem_prepared:
            shutil.rmtree(self._tmp_dir)
        TaurusBaseEditor.closeEvent(self, event)


def demo(model_name="MS_BL98"):
    test = SardanaEditor()
    test.resize(1000, 800)
    # test.load(__file__)
    test.setModel(model_name)
    test.show()
    return test


def main():
    import sys
    import taurus.qt.qtgui.application
    Application = taurus.qt.qtgui.application.TaurusApplication

    app = Application.instance()
    owns_app = app is None

    if owns_app:
        import taurus.core.util.argparse
        parser = taurus.core.util.argparse.get_taurus_parser()
        parser.usage = "%prog [options] <macro server name>"

        app = Application(sys.argv, cmd_line_parser=parser,
                          app_name="Macro editor demo", app_version="1.0",
                          org_domain="Sardana", org_name="Tango community")

    args = app.get_command_line_args()

    if len(args) == 0:
        parser.error("must give a macroserver device name")
    else:
        w = demo(args[0])

    if owns_app:
        sys.exit(app.exec_())
    else:
        return w

if __name__ == "__main__":
    main()



#!/usr/bin/env python

##############################################################################
##
# This file is part of Sardana
##
# http://www.sardana-controls.org/
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""This module is part of the Python Pool library. It defines the base classes
for TwoDExpChannel"""

__all__ = ["Pool2DExpChannel"]

__docformat__ = 'restructuredtext'

from sardana import ElementType
from sardana.sardanaevent import EventType

from sardana.pool.poolbasechannel import PoolTimerableChannel


class Pool2DExpChannel(PoolTimerableChannel):

    def __init__(self, **kwargs):
        self._data_source = None
        self._timer = None
        kwargs['elem_type'] = ElementType.TwoDExpChannel
        PoolTimerableChannel.__init__(self, **kwargs)

    # --------------------------------------------------------------------------
    # data source
    # --------------------------------------------------------------------------

    def get_data_source(self, cache=True, propagate=1):
        msg = ("data_source is deprecated since "
               "version 2.8.0. Use value_ref instead.")
        self.warning(msg)
        if not cache or self._data_source is None:
            data_source = self.read_data_source()
            self._set_data_source(data_source, propagate=propagate)
        return self._data_source

    def _set_data_source(self, data_source, propagate=1):
        self._data_source = data_source
        if not propagate:
            return
        self.fire_event(
            EventType("data_source", priority=propagate), data_source)

    def read_data_source(self):
        data_source = self.controller.get_axis_par(self.axis, "data_source")
        return data_source

    data_source = property(
        get_data_source, doc="source identifier for the 2D data")



#!/usr/bin/env python

import sys
import types
from lxml import etree


def transform(f):
    t = etree.XSLT(etree.parse("SAR_TO_FODS.xslt"))
    if isinstance(f, str):
        doc = etree.parse(f)
    else:
        doc = f
    return t(doc)


def main():
    filename = sys.argv[1]
    t = transform(filename)
    print(etree.tostring(t, pretty_print=True))

if __name__ == "__main__":
    main()



#!/usr/bin/env python

##############################################################################
##
# This file is part of Sardana
##
# http://www.tango-controls.org/static/sardana/latest/doc/html/index.html
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################
import threading

import unittest

from sardana.pool.poolsynchronization import PoolSynchronization
from sardana.pool.poolacquisition import get_acq_ctrls
from sardana.sardanadefs import State
from sardana.pool.test import FakePool, createPoolController, \
    createPoolTriggerGate, dummyPoolTGCtrlConf01, dummyTriggerGateConf01, \
    createControllerConfiguration


class PoolTriggerGateTestCase(unittest.TestCase):
    """Unittest of PoolSynchronization class"""

    def setUp(self):
        """Create a Controller, TriggerGate and PoolSynchronization objects from
        dummy configurations
        """
        unittest.TestCase.setUp(self)
        try:
            from mock import Mock
        except ImportError:
            self.skipTest("mock module is not available")
        pool = FakePool()
        dummy_tg_ctrl = createPoolController(pool, dummyPoolTGCtrlConf01)
        self.dummy_tg = createPoolTriggerGate(pool, dummy_tg_ctrl,
                                              dummyTriggerGateConf01)
        dummy_tg_ctrl.add_element(self.dummy_tg)
        pool.add_element(dummy_tg_ctrl)
        pool.add_element(self.dummy_tg)
        self.conf_ctrl = createControllerConfiguration(dummy_tg_ctrl,
                                                       [self.dummy_tg])

        self.ctrls = get_acq_ctrls([self.conf_ctrl])
        # self.cfg = createPoolSynchronizationConfiguration((dummy_tg_ctrl,),
        #                                                   ((self.dummy_tg,),),)
        # Create mock and define its functions
        ctrl_methods = ['PreStartAll', 'StartAll', 'PreStartOne', 'StartOne',
                        'PreStateAll', 'StateAll', 'PreStateOne', 'StateOne',
                        'PreSynchAll', 'PreSynchOne', 'SynchOne', 'SynchAll']
        self.mock_tg_ctrl = Mock(spec=ctrl_methods)
        self.mock_tg_ctrl.StateOne.return_value = (State.Moving, 'triggering')

        dummy_tg_ctrl.ctrl = self.mock_tg_ctrl
        self.tgaction = PoolSynchronization(self.dummy_tg)
        self.tgaction.add_element(self.dummy_tg)

    def stopGeneration(self):
        """Method used to change the controller (mock) state"""
        self.mock_tg_ctrl.StateOne.return_value = (State.On, 'On')

    def test_tggeneration(self):
        """Verify trigger element states before and after action_loop."""
        from mock import call, MagicMock
        # starting action
        synchronization = MagicMock()
        self.tgaction.start_action(self.ctrls, synchronization)
        # verifying that the action correctly started the involved controller
        self.mock_tg_ctrl.assert_has_calls([call.PreStartAll(),
                                            (call.PreStartOne(1,)),
                                            (call.StartOne(1,)),
                                            (call.StartAll())])
        # verifying that the elements involved in action changed its state
        element_state = self.dummy_tg.get_state()
        msg = ("State after start_action is '%s'. (Expected: '%s')" %
               (State.get(element_state), "Moving"))
        self.assertEqual(element_state, State.Moving, msg)
        # starting timer (1 s) which will change the controller state
        threading.Timer(1, self.stopGeneration).start()
        # entering action loop
        self.tgaction.action_loop()
        # verifying that the action checked the controller states
        self.mock_tg_ctrl.assert_has_calls([call.PreStateAll(),
                                            call.PreStateOne(1,),
                                            call.StateAll(),
                                            call.StateOne(1,)])
        # verifying that the elements involved in action changed its state
        element_state = self.dummy_tg.get_state()
        msg = ("State after action_loop shall be different than Moving")
        self.assertNotEqual(element_state, State.Moving, msg)

    def tearDown(self):
        self.tgaction = None
        self.mock_tg_ctrl = None
        self.cfg = None
        self.dummy_tg = None
        unittest.TestCase.tearDown(self)



import uuid
import unittest
from taurus import Device
from taurus.core.tango.tangovalidator import TangoDeviceNameValidator
from sardana.pool import AcqSynchType
from sardana.taurus.core.tango.sardana.pool import registerExtensions
from sardana.tango.pool.test.base_sartest import SarTestTestCase


class TestMeasurementGroupConfiguration(SarTestTestCase, unittest.TestCase):

    def setUp(self):
        SarTestTestCase.setUp(self)
        registerExtensions()

    def tearDown(self):
        SarTestTestCase.tearDown(self)

    def _assertResult(self, result, channels, expected_value):
        expected_channels = list(channels)
        for channel, value in result.items():
            msg = "unexpected key: {}".format(channel)
            self.assertIn(channel, expected_channels, msg)
            expected_channels.remove(channel)
            self.assertEqual(value, expected_value)
        msg = "{} are missing".format(expected_channels)
        self.assertEqual(len(expected_channels), 0, msg)

    def _assertMultipleResults(self, result, channels, expected_values):
        expected_channels = list(channels)
        for (channel, value), expected_value in zip(result.items(),
                                                    expected_values):
            msg = "unexpected key: {}".format(channel)
            self.assertIn(channel, expected_channels, msg)
            expected_channels.remove(channel)
            self.assertEqual(value, expected_value)
        msg = "{} are missing".format(expected_channels)
        self.assertEqual(len(expected_channels), 0, msg)

    def test_enabled(self, elements=["_test_ct_1_1", "_test_ct_1_2",
                                     "_test_2d_1_3",
                                     "_test_mt_1_3/position"]):
        mg_name = str(uuid.uuid1())
        argin = [mg_name] + elements
        self.pool.CreateMeasurementGroup(argin)
        try:
            mg = Device(mg_name)

            # Check initial state of all kind of channels, nonexistent
            # channels for the feature return None as result.
            result = mg.getEnabled(*elements)
            expected = [True] * len(elements)
            self._assertMultipleResults(result, elements, expected)

            # Test every possible combination of setting values
            # Check that changing one channel doesn't affect the other
            mg.setEnabled(False, *elements)
            result = mg.getEnabled(*elements)
            expected = [False] * len(elements)
            self._assertMultipleResults(result, elements, expected)
            mg.setEnabled(True, elements[0])
            result = mg.getEnabled(*elements)
            expected = [False] * len(elements)
            expected[0] = True
            self._assertMultipleResults(result, elements, expected)
            mg.setEnabled(False, *elements)
            resutl = mg.getEnabled(*elements)
            self._assertResult(resutl, elements, False)

            # Redefine elements to ony use existing values
            elements = ["_test_ct_1_1", "_test_ct_1_2"]

            # Set values using the controller instead of channels
            mg.setEnabled(True, "_test_ct_ctrl_1")
            resutl = mg.getEnabled(*elements)
            self._assertResult(resutl, elements, True)

            # Get values by controller
            mg.setEnabled(False, *elements)
            resutl = mg.getEnabled("_test_ct_ctrl_1")
            self._assertResult(resutl, elements, False)

            # Check ret_full_name
            v = TangoDeviceNameValidator()
            full_names = [v.getNames(element)[0] for element in elements]
            resutl = mg.getEnabled(*full_names)
            self._assertResult(resutl, elements, False)
            mg.setEnabled(True, *full_names)
            resutl = mg.getEnabled(*elements, ret_full_name=True)
            self._assertResult(resutl, full_names, True)
        finally:
            mg.cleanUp()
            self.pool.DeleteElement(mg_name)

    def test_output(self, elements=["_test_ct_1_1", "_test_ct_1_2",
                                    "_test_2d_1_3",
                                    "_test_mt_1_3/position"]):
        mg_name = str(uuid.uuid1())
        argin = [mg_name] + elements
        self.pool.CreateMeasurementGroup(argin)

        try:
            mg = Device(mg_name)

            # Check initial state of all kind of channels, nonexistent
            # channels for the feature return None as result.
            enabled = mg.getOutput(*elements)
            expected = [True] * len(elements)
            self._assertMultipleResults(enabled, elements, expected)

            # Test every possible combination of setting values
            # Check that changing one channel doesn't affect the other
            mg.setOutput(False, *elements)
            is_output = mg.getOutput(*elements)
            self._assertResult(is_output, elements, False)
            mg.setOutput(True, elements[0])
            result = mg.getOutput(*elements)
            expected = [False] * len(elements)
            expected[0] = True
            self._assertMultipleResults(result, elements, expected)
            mg.setOutput(False, *elements)
            is_output = mg.getOutput(*elements)
            self._assertResult(is_output, elements, False)

            # Redefine elements to ony use existing values
            elements = ["_test_ct_1_1", "_test_ct_1_2"]

            # Set values using the controller instead of channels
            mg.setOutput(True, "_test_ct_ctrl_1")
            is_output = mg.getOutput(*elements)
            self._assertResult(is_output, elements, True)

            # Get values by controller
            mg.setOutput(False, *elements)
            is_output = mg.getOutput("_test_ct_ctrl_1")
            self._assertResult(is_output, elements, False)

            # Check ret_full_name
            v = TangoDeviceNameValidator()
            full_names = [v.getNames(element)[0] for element in elements]
            is_output = mg.getOutput(*full_names)
            self._assertResult(is_output, elements, False)
            mg.setOutput(True, *full_names)
            is_output = mg.getOutput(*elements, ret_full_name=True)
            self._assertResult(is_output, full_names, True)
        finally:
            mg.cleanUp()
            self.pool.DeleteElement(mg_name)

    def test_PlotType(self, elements=["_test_ct_1_1", "_test_ct_1_2",
                                      "_test_ct_1_3", "_test_2d_1_3",
                                      "_test_mt_1_3/position"]):
        mg_name = str(uuid.uuid1())
        argin = [mg_name] + elements
        self.pool.CreateMeasurementGroup(argin)

        try:
            mg = Device(mg_name)
            plottype = mg.getPlotType()
            self._assertResult(plottype, elements, 0)
            mg.setPlotType("Image", elements[0])
            mg.setPlotType("Spectrum", elements[1])
            mg.setPlotType("No", elements[2])
            mg.setPlotType("Image", elements[3])
            mg.setPlotType("Image", elements[4])
            plottype = mg.getPlotType()
            expected_values = [2, 1, 0, 2, 2]
            self._assertMultipleResults(plottype, elements, expected_values)
            with self.assertRaises(ValueError):
                mg.setPlotType("asdf", elements[2])

            # Redefine elements
            elements = ["_test_ct_1_1", "_test_ct_1_2", "_test_ct_1_3"]

            # Set values using the controller instead of channels
            mg.setPlotType("Image", "_test_ct_ctrl_1")
            plottype = mg.getPlotType(*elements)
            self._assertResult(plottype, elements, 2)

            # Get values by controller
            mg.setPlotType("Spectrum", *elements)
            plottype = mg.getPlotType("_test_ct_ctrl_1")
            self._assertResult(plottype, elements, 1)

            # Check ret_full_name
            v = TangoDeviceNameValidator()
            full_names = [v.getNames(element)[0] for element in elements]
            plottype = mg.getPlotType(*full_names)
            self._assertResult(plottype, elements, 1)
            mg.setPlotType("Image", *full_names)
            plottype = mg.getPlotType(*elements, ret_full_name=True)
            self._assertResult(plottype, full_names, 2)

        finally:
            mg.cleanUp()
            self.pool.DeleteElement(mg_name)

    def test_PlotAxes(self, elements=["_test_ct_1_1", "_test_ct_1_2",
                                      "_test_ct_1_3", "_test_2d_1_3",
                                      "_test_mt_1_3/position"]):
        mg_name = str(uuid.uuid1())
        argin = [mg_name] + elements
        self.pool.CreateMeasurementGroup(argin)

        try:
            mg = Device(mg_name)
            mg.setPlotType("Image", elements[0])
            mg.setPlotType("Spectrum", elements[1])
            mg.setPlotType("No", elements[2])
            mg.setPlotType("Image", elements[3])
            mg.setPlotType("Image", elements[4])
            result = mg.getPlotAxes()
            self._assertResult(result, elements, [])

            mg.setPlotAxes(["<idx>", "<idx>"], elements[0])
            mg.setPlotAxes(["<mov>"], elements[1])
            with self.assertRaises(Exception):
                mg.setPlotAxes(['<mov>'], elements[2])
            mg.setPlotAxes(["<idx>", "<idx>"], elements[3])
            mg.setPlotAxes(["<idx>", "<idx>"], elements[4])

            result = mg.getPlotAxes()
            expected_result = [['<idx>', '<idx>'], ['<mov>'], [],
                               ['<idx>', '<idx>'], ['<idx>', '<idx>']]
            self._assertMultipleResults(result, elements, expected_result)

            mg.setPlotAxes(["<mov>", "<idx>"], elements[0])
            mg.setPlotAxes(["<idx>"], elements[1])
            result = mg.getPlotAxes()
            expected_result = [['<mov>', '<idx>'], ['<idx>'], [],
                               ['<idx>', '<idx>'], ['<idx>', '<idx>']]
            self._assertMultipleResults(result, elements, expected_result)

            mg.setPlotAxes(["<mov>", "<mov>"], elements[0])
            result = mg.getPlotAxes()
            expected_result = [['<mov>', '<mov>'], ['<idx>'], [],
                               ['<idx>', '<idx>'], ['<idx>', '<idx>']]
            self._assertMultipleResults(result, elements, expected_result)

            with self.assertRaises(RuntimeError):
                mg.setPlotAxes(["<mov>"], elements[2])
            with self.assertRaises(ValueError):
                mg.setPlotAxes(["<mov>", "<idx>"], elements[1])
            with self.assertRaises(ValueError):
                mg.setPlotAxes(["<mov>"], elements[0])

            elements = ["_test_ct_1_1", "_test_ct_1_2", "_test_ct_1_3"]
            # Set values using the controller instead of channels
            with self.assertRaises(Exception):
                mg.setPlotAxes(["<mov>"], "_test_ct_ctrl_1")
            # TODO get method by controller doesn't give the order
            # Get values by controller
            result = mg.getPlotAxes("_test_ct_ctrl_1")
            expected_result = [['<mov>', '<mov>'], ['<idx>'], []]
            self._assertMultipleResults(result, elements, expected_result)

            # Check ret_full_name
            v = TangoDeviceNameValidator()
            full_names = [v.getNames(element)[0] for element in elements]
            result = mg.getPlotAxes(*full_names)
            expected_result = [['<mov>', '<mov>'], ['<idx>'], []]
            self._assertMultipleResults(result, elements, expected_result)
            mg.setPlotAxes(["<idx>", "<idx>"], full_names[0])
            mg.setPlotAxes(["<mov>"], full_names[1])
            result = mg.getPlotAxes(*elements, ret_full_name=True)
            expected_result = [['<idx>', '<idx>'], ['<mov>'], []]
            self._assertMultipleResults(result, full_names, expected_result)
        finally:
            mg.cleanUp()
            self.pool.DeleteElement(mg_name)

    def test_Timer(self, elements=["_test_ct_1_1", "_test_ct_1_2",
                                   "_test_ct_1_3",
                                   "_test_mt_1_3/position"]):
        mg_name = str(uuid.uuid1())
        argin = [mg_name] + elements
        self.pool.CreateMeasurementGroup(argin)
        try:
            mg = Device(mg_name)

            result = mg.getTimer("_test_mt_1_3/position")
            with self.assertRaises(Exception):
                mg.setTimer("_test_mt_1_3/position")
            self._assertResult(result,  ["_test_mt_1_3/position"], None)
            mg.setTimer('_test_ct_1_3')
            result = mg.getTimer(*elements)
            expected = ['_test_ct_1_3', '_test_ct_1_3', '_test_ct_1_3', None]
            self._assertMultipleResults(result, elements, expected)

            mg.setTimer('_test_ct_1_2')
            result = mg.getTimer(*elements)
            expected = ['_test_ct_1_2', '_test_ct_1_2', '_test_ct_1_2', None]
            self._assertMultipleResults(result, elements, expected)

            result = mg.getTimer(*elements, ret_by_ctrl=True)
            self._assertMultipleResults(result,
                                        ['_test_ct_ctrl_1', '__tango__'],
                                        ['_test_ct_1_2', None])

            # Check ret_full_name
            v = TangoDeviceNameValidator()
            counters = ["_test_ct_1_1", "_test_ct_1_2", "_test_ct_1_3"]
            full_names = [v.getNames(counter)[0] for counter in counters]
            mg.setTimer(v.getNames('_test_ct_1_1')[0])
            result = mg.getTimer()
            expected = ['_test_ct_1_1', '_test_ct_1_1', '_test_ct_1_1', None]
            self._assertMultipleResults(result, elements, expected)
            # TODO ret_full_name gives controler name
            mg.setTimer("_test_ct_1_2")
            result = mg.getTimer(*counters, ret_full_name=True)
            self._assertResult(result, full_names, "_test_ct_1_2")
        finally:
            mg.cleanUp()
            self.pool.DeleteElement(mg_name)

    def test_Monitor(self, elements=["_test_ct_1_1", "_test_ct_1_2",
                                     "_test_ct_1_3", "_test_2d_1_1",
                                     '_test_2d_1_2',
                                     "_test_mt_1_3/position"]):
        mg_name = str(uuid.uuid1())
        argin = [mg_name] + elements
        self.pool.CreateMeasurementGroup(argin)
        try:
            mg = Device(mg_name)

            with self.assertRaises(Exception):
                mg.setMonitor("_test_mt_1_3/position")

            mg.setMonitor('_test_2d_1_2')
            mg.setMonitor("_test_ct_1_3")
            expected = ["_test_ct_1_3", "_test_ct_1_3", "_test_ct_1_3",
                        "_test_2d_1_2", '_test_2d_1_2', None]
            result = mg.getMonitor()
            self._assertMultipleResults(result, elements, expected)

            expected = ["_test_ct_1_3", '_test_2d_1_2', None]
            result = mg.getMonitor(ret_by_ctrl=True)
            ctrls = ['_test_ct_ctrl_1', '_test_2d_ctrl_1', '__tango__']
            self._assertMultipleResults(result, ctrls, expected)

            # Check ret_full_name
            v = TangoDeviceNameValidator()
            counters = ["_test_ct_1_1", "_test_ct_1_2", "_test_ct_1_3",
                        '_test_2d_1_1', '_test_2d_1_2']
            full_names = [v.getNames(counter)[0] for counter in counters]
            mg.setMonitor(v.getNames('_test_ct_1_1')[0])
            mg.setMonitor(v.getNames('_test_2d_1_2')[0])

            result = mg.getMonitor(*counters, ret_full_name=True)
            expected = ["_test_ct_1_1", "_test_ct_1_1", "_test_ct_1_1",
                        "_test_2d_1_2", '_test_2d_1_2']
            self._assertMultipleResults(result, full_names, expected)
        finally:
            mg.cleanUp()
            self.pool.DeleteElement(mg_name)

    def test_Synchronizer(self, elements=["_test_ct_1_1", "_test_ct_1_2",
                                          "_test_ct_1_3", "_test_2d_1_1",
                                          "_test_mt_1_3/position"]):
        mg_name = str(uuid.uuid1())
        argin = [mg_name] + elements
        self.pool.CreateMeasurementGroup(argin)
        try:
            mg = Device(mg_name)
            result = mg.getSynchronizer()
            expected = ['software', 'software', 'software', 'software', None]
            self._assertMultipleResults(result, elements, expected)
            with self.assertRaises(Exception):
                mg.setSynchronizer('_test_tg_1_2', "_test_mt_1_3/position")

            mg.setSynchronizer('_test_tg_1_2', "_test_ct_ctrl_1",
                               "_test_2d_ctrl_1")

            expected = ['_test_tg_1_2', '_test_tg_1_2', '_test_tg_1_2',
                        '_test_tg_1_2', None]
            result = mg.getSynchronizer()
            self._assertMultipleResults(result, elements, expected)

            mg.setSynchronizer('software', "_test_ct_ctrl_1",
                               "_test_2d_ctrl_1")
            result = mg.getSynchronizer()
            expected = ['software', 'software', 'software', 'software', None]
            self._assertMultipleResults(result, elements, expected)

            with self.assertRaises(Exception):
                mg.setSynchronizer('asdf', "_test_ct_ctrl_1",
                                   "_test_2d_ctrl_1")

            # Check ret_full_name
            v = TangoDeviceNameValidator()
            counters = ["_test_ct_1_1", "_test_ct_1_2", "_test_ct_1_3",
                        '_test_2d_1_1']
            full_names = [v.getNames(counter)[0] for counter in counters]
            mg.setSynchronizer('_test_tg_1_2', "_test_ct_ctrl_1",
                               "_test_2d_ctrl_1")

            result = mg.getSynchronizer(*counters, ret_full_name=True)

            self._assertResult(result, full_names, '_test_tg_1_2')

        finally:
            mg.cleanUp()
            self.pool.DeleteElement(mg_name)

    def test_Synchronization(self, elements=["_test_ct_1_1", "_test_ct_1_2",
                                             "_test_ct_1_3", "_test_2d_1_1",
                                             "_test_mt_1_3/position"]):
        mg_name = str(uuid.uuid1())
        argin = [mg_name] + elements
        self.pool.CreateMeasurementGroup(argin)
        try:
            mg = Device(mg_name)
            result = mg.getSynchronization()
            expected = [AcqSynchType.Trigger, AcqSynchType.Trigger,
                        AcqSynchType.Trigger, AcqSynchType.Trigger, None]
            self._assertMultipleResults(result, elements, expected)
            # TODO: maybe we should raise an exception here?
            # with self.assertRaises(Exception):
            #     mg.setSynchronization(AcqSynchType.Trigger,
            #                           "_test_mt_1_3/position")

            mg.setSynchronization(AcqSynchType.Gate, "_test_ct_ctrl_1",
                                  "_test_2d_ctrl_1")

            expected = [AcqSynchType.Gate, AcqSynchType.Gate,
                        AcqSynchType.Gate, AcqSynchType.Gate, None]
            result = mg.getSynchronization()
            self._assertMultipleResults(result, elements, expected)

            mg.setSynchronization(AcqSynchType.Start, "_test_ct_ctrl_1",
                                  "_test_2d_ctrl_1")
            result = mg.getSynchronization()
            expected = [AcqSynchType.Start, AcqSynchType.Start,
                        AcqSynchType.Start, AcqSynchType.Start, None]
            self._assertMultipleResults(result, elements, expected)

            with self.assertRaises(Exception):
                mg.setSynchronization('asdf', "_test_ct_ctrl_1",
                                      "_test_2d_ctrl_1")

            # Check ret_full_name
            v = TangoDeviceNameValidator()
            counters = ["_test_ct_1_1", "_test_ct_1_2", "_test_ct_1_3",
                        '_test_2d_1_1']
            full_names = [v.getNames(counter)[0] for counter in counters]
            mg.setSynchronization(AcqSynchType.Trigger, "_test_ct_ctrl_1",
                                  "_test_2d_ctrl_1")

            result = mg.getSynchronization(*counters, ret_full_name=True)

            self._assertResult(result, full_names, AcqSynchType.Trigger)

        finally:
            mg.cleanUp()
            self.pool.DeleteElement(mg_name)

    def test_ValueRefEnabled(self, elements=["_test_2d_1_1", "_test_2d_1_2",
                                             "_test_ct_1_3",
                                             "_test_mt_1_3/position"]):
        mg_name = str(uuid.uuid1())
        argin = [mg_name] + elements
        self.pool.CreateMeasurementGroup(argin)
        try:
            mg = Device(mg_name)

            # Check initial state of all kind of channels, nonexistent
            # channels for the feature return None as result.
            enabled = mg.getValueRefEnabled(*elements)
            expected = [False, False, None, None]
            self._assertMultipleResults(enabled, elements, expected)

            # Check if the nonexistent channels raise error if trying to set
            with self.assertRaises(Exception):
                mg.setValueRefEnabled(True, *elements[-2])
            with self.assertRaises(Exception):
                mg.setValueRefEnabled(True, *elements[-1])

            # Redefine elements to ony use existing values
            elements = ["_test_2d_1_1", "_test_2d_1_2"]

            # Test every possible combination of setting values
            # Check that changing one channel doesn't affect the other
            mg.setValueRefEnabled(True, *elements)
            enabled = mg.getValueRefEnabled(*elements)
            self._assertResult(enabled, elements, True)
            mg.setValueRefEnabled(False, elements[0])
            result = mg.getValueRefEnabled(*elements)
            expected = [True] * len(elements)
            expected[0] = False
            self._assertMultipleResults(result, elements, expected)
            mg.setValueRefEnabled(True, *elements)
            enabled = mg.getValueRefEnabled(*elements)
            self._assertResult(enabled, elements, True)

            # Set values using the controller instead of channels
            mg.setValueRefEnabled(True, "_test_2d_ctrl_1")
            enabled = mg.getValueRefEnabled(*elements)
            self._assertResult(enabled, elements, True)

            # Get values by controller
            mg.setValueRefEnabled(False, *elements)
            enabled = mg.getValueRefEnabled("_test_2d_ctrl_1")
            self._assertResult(enabled, elements, False)

            # Check ret_full_name
            v = TangoDeviceNameValidator()
            full_names = [v.getNames(element)[0] for element in elements]
            enabled = mg.getValueRefEnabled(*full_names)
            self._assertResult(enabled, elements, False)
            mg.setValueRefEnabled(True, *full_names)
            enabled = mg.getValueRefEnabled(*elements, ret_full_name=True)
            self._assertResult(enabled, full_names, True)
        finally:
            mg.cleanUp()
            self.pool.DeleteElement(mg_name)

    def test_ValueRefPattern(self, elements=["_test_2d_1_1", "_test_2d_1_2",
                                             "_test_ct_1_3",
                                             "_test_mt_1_3/position"]):
        mg_name = str(uuid.uuid1())
        argin = [mg_name] + elements
        self.pool.CreateMeasurementGroup(argin)
        try:
            mg = Device(mg_name)

            # Check initial state of all kind of channels, nonexistent
            # channels for the feature return None as result.
            pattern = mg.getValueRefPattern(*elements)
            expected = ['', '', None, None]
            self._assertMultipleResults(pattern, elements, expected)

            # Check if the nonexistent channels raise error if trying to set
            with self.assertRaises(Exception):
                mg.setValueRefPattern('/tmp/test_foo.txt', *elements[-2])
            with self.assertRaises(Exception):
                mg.setValueRefPattern('/tmp/test_foo.txt', *elements[-1])

            # Redefine elements to ony use existing values
            elements = ["_test_2d_1_1", "_test_2d_1_2"]

            # Test every possible combination of setting values
            # Check that changing one channel doesn't affect the other
            mg.setValueRefPattern('/tmp/test_foo.txt', *elements)
            pattern = mg.getValueRefPattern(*elements)
            self._assertResult(pattern, elements, '/tmp/test_foo.txt')

            # Set values using the controller instead of channels
            mg.setValueRefPattern('/tmp/test_foo2.txt', "_test_2d_ctrl_1")
            pattern = mg.getValueRefPattern(*elements)
            self._assertResult(pattern, elements, '/tmp/test_foo2.txt')

            # Get values by controller
            mg.setValueRefPattern('/tmp/test_foo.txt', *elements)
            pattern = mg.getValueRefPattern("_test_2d_ctrl_1")
            self._assertResult(pattern, elements, '/tmp/test_foo.txt')

            # Check ret_full_name
            v = TangoDeviceNameValidator()
            full_names = [v.getNames(element)[0] for element in elements]
            pattern = mg.getValueRefPattern(*full_names)
            self._assertResult(pattern, elements, '/tmp/test_foo.txt')
            mg.setValueRefPattern('/tmp/test_foo2.txt', *full_names)
            pattern = mg.getValueRefPattern(*elements, ret_full_name=True)
            self._assertResult(pattern, full_names, '/tmp/test_foo2.txt')

        finally:
            mg.cleanUp()
            self.pool.DeleteElement(mg_name)



##############################################################################
##
# This file is part of Sardana
##
# http://www.sardana-controls.org/
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""Library of specific macros for performing specific experimental techniques

"""

__all__ = ["xas_acq"]

__docformat__ = 'restructuredtext'

import numpy

from sardana.macroserver.macro import *
from sardana.macroserver.scan import *


class xas_acq(Macro, Hookable):
    """
    .. warning:: This macro is still under development. It may change.

    Perform an X-ray absorption scan experiment. Data is stored in a NXxas-compliant file.
    """
    hints = {'scan': 'xas_acq', 'allowsHooks': (
        'pre-move', 'post-move', 'pre-acq', 'post-acq', 'post-step')}
    # env = ('MonochromatorEnergy', )#'AbsorbedBeam', 'IncomingBeam',
    # 'Monitor') #this hints that the macro requires the ActiveMntGrp
    # environment variable to be set

    param_def = [
        ['start',  Type.Float,    None, 'start energy in keV'],
        ['final',  Type.Float,    None, 'final energy in keV'],
        ['nr_interv',  Type.Integer,  None, 'Number of energy intervals'],
        ['integ_time', Type.Float,    None, 'Integration time in s']
    ]

    def prepare(self, start, final, nr_interv, integ_time, **opts):
        # parse the user parameters
        self.starts = numpy.array([start], dtype='d')
        self.finals = numpy.array([final], dtype='d')
        self.integ_time = integ_time

        self.nb_points = nr_interv + 1
        self.interv_sizes = (self.finals - self.starts) / nr_interv
        self.name = 'xas_acq'
        # the "env" dictionary may be passed as an option
        env = opts.get('env', {})
        env['integ_time'] = integ_time

        # print "!!!!!", type(self.getInstrument('/instrument/monochromator')), self.getEnv('MonochromatorEnergy', macro_name=self.name)
        # ElementWithInterface('Instrument','monochromator')
        for n, e in self.getElementsWithInterface('Instrument').items():
            inst = e.getObj()
            # ,inst.getElements()
            print(n, e.name, inst.getFullName(), type(e), type(inst),
                  type(inst.getPoolObj()))

        # maybe I should use the instrument interface to obtain the right
        # counters

        env['monochromator'] = monochromator = self.getEnv(
            'MonochromatorEnergy', macro_name=self.name)
        energymotor = self.getMoveable(monochromator)
        xasMntGrp = self.getMeasurementGroup(
            self.getEnv('xasMntGrp', macro_name=self.name))

        xasMntGrp = xasMntGrp.getObj()
        monitor, incbeam, absbeam = xasMntGrp.getChannelNames()[:3]
        env['monitor'] = monitor
        env['absbeam'] = absbeam
        env['incbeam'] = incbeam

#        print "!!!!!!!!",xasMntGrp.getObj().getElements()


#        absbeam =  self.getExpChannel(self.getEnv('AbsorbedBeam')) #this should be get measurement group (e.g., second channel?)
#        incbeam =  self.getExpChannel(self.getEnv('IncomingBeam')) #this should be get measurement group (e.g., first channel?)
# monitor =  self.getExpChannel(self.getEnv('Monitor')) #this should be
# get from the monitor of the measurement group

        # create an instance of GScan (in this case, of its child, SScan
        self._gScan = SScan(self, generator=self._generator,
                            moveables=[energymotor], env=env)

    def _generator(self):
        step = {}
        step["integ_time"] = self.integ_time
        step[
            "post-acq-hooks"] = self.getHooks('post-acq') + self.getHooks('_NOHINT_')
        step["post-step-hooks"] = self.getHooks('post-step')
        step["check_func"] = []
        for point_no in range(self.nb_points):
            step["positions"] = self.starts + point_no * self.interv_sizes
            step["point_id"] = point_no
            yield step

    def run(self, *args):
        for step in self._gScan.step_scan():  # just go through the steps
            yield step

    @property
    def data(self):
        return self._gScan.data  # the GScan provides scan data

    def _get_nr_points(self):
        msg = ("nr_points is deprecated since version 3.0.3. "
               "Use nb_points instead.")
        self.warning(msg)
        return self.nb_points

    nr_points = property(_get_nr_points)



##############################################################################
##
# This file is part of Sardana
##
# http://www.sardana-controls.org/
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""MCA related macros"""

__docformat__ = 'restructuredtext'

__all__ = ["mca_start", "mca_stop"]

from sardana.macroserver.macro import *

import array


class mca_start(Macro):
    """Starts an mca"""

    param_def = [
        ['mca', Type.ExpChannel, None, 'mca']
    ]

    def run(self, mca):
        name = mca.getName()
        mca.Start()
        o = "Starting " + name
        self.output(o)


class mca_stop(Macro):
    """Stops an mca"""

    param_def = [
        ['mca', Type.ExpChannel, None, 'mca']
    ]

    def run(self, mca):
        name = mca.getName()
        mca.Abort()
        o = "Stopping " + name
        self.output(o)



"""
Create a stand-alone ScanData object to which we can call addRecord
Do what gscan does for creating the ScanData, without using gscan.

- generate a minimum ScanDataEnvironment
- create a datahandler with a NXScan recorder only
- check if we can call addRecord and that a record is written in the recorder
- Check that we can call addData
"""

__all__ = ['createScanDataEnvironment', 'DummyEventSource']


import time
from datetime import date
import threading
import numpy
import os

from sardana.macroserver.scan.recorder import DataHandler
from sardana.macroserver.recorders.storage import NXscan_FileRecorder

from sardana.macroserver.scan.scandata import (ScanData, ScanDataEnvironment,
                                               ColumnDesc)


class DummyEventSource(threading.Thread):

    def __init__(self, name, scanData, values, intervals=None):
        threading.Thread.__init__(self, name=name)
        self.scan_data = scanData
        self.values = values
        self.intervals = intervals or numpy.random.rand(len(self.values))

    def run(self):
        i = 0
        for v, t in zip(self.values, self.intervals):
            try:
                idx = list(range(i, i + len(v)))
                i += len(v)
                skip = float('NaN') in v
            except TypeError:  # if v is not a list
                idx = [i]
                i += 1
                v = [v]
                skip = float('NaN') in v
            if skip:
                continue
            time.sleep(t)
            _dict = dict(data=v, index=idx, label=self.name)
            self.scan_data.addData(_dict)

    def get_obj(self):
        return self


def createScanDataEnvironment(columns, scanDir='/tmp/',
                              scanFile='data_nxs.hdf5'):

    serialno = 1
    env = ScanDataEnvironment(
        {'serialno': serialno,
         'user': "USER_NAME",
         'title': "title_macro"})

    env['name'] = "env_name"
    env['estimatedtime'] = -1.0
    env['ScanDir'] = scanDir
    env['ScanFile'] = scanFile
    env['total_scan_intervals'] = -1.0

    today = date.today()
    env['datetime'] = today
    env['starttime'] = today
    env['endtime'] = today

    # Initialize the data_desc list (and add the point number column)
    data_desc = []
    column_master = ColumnDesc(name='point_nb',
                               label='#Pt No', dtype='int64')
    data_desc.append(column_master)

    for i in range(len(columns)):
        col = columns[i]
        column = ColumnDesc(name=col, label=col, dtype='float64')
        data_desc.append(column)

    env['datadesc'] = data_desc
    return env


def main():
    data_handler = DataHandler()
    file_name = "/tmp/data_nxs.hdf5"
    nx_recorder = NXscan_FileRecorder(filename=file_name,
                                      macro="dscan", overwrite=True)
    data_handler.addRecorder(nx_recorder)
    scan_dir, scan_file = os.path.split(file_name)
    columns = ['ch1', 'ch2']

    env = createScanDataEnvironment(columns, scan_dir, scan_file)
    scan_data = ScanData(environment=env, data_handler=data_handler)

    data1 = [10.0, [6.0, 3.4]]
    data2 = [10.0, None, None, 5]

    srcs = [
        DummyEventSource(columns[0], scan_data, data1),
        DummyEventSource(columns[1], scan_data, data2)
    ]
    scan_data.start()
    for s in srcs:
        s.start()
    for s in srcs:
        s.join()
    scan_data.end()
    # Test read nxs file
    import nxs
    f = nxs.load(file_name)
    m = f['entry1']['measurement']
    ch1 = m['ch1']
    print(ch1.nxdata)

if __name__ == "__main__":
    main()
